# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements OpenGL lookAt functionalities."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.math import vector
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def right_handed(camera_position, look_at, up_vector, name=None):
  """Builds a right handed look at view matrix.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    camera_position: A tensor of shape `[A1, ..., An, 3]`, where the last
      dimension represents the 3D position of the camera.
    look_at: A tensor of shape `[A1, ..., An, 3]`, with the last dimension
      storing the position where the camera is looking at.
    up_vector: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      defines the up vector of the camera.
    name: A name for this op. Defaults to 'right_handed'.

  Raises:
    ValueError: if the all the inputs are not of the same shape, or if any input
    of of an unsupported shape.

  Returns:
    A tensor of shape `[A1, ..., An, 4, 4]`, containing right handed look at
    matrices.
  """
  with tf.compat.v1.name_scope(name, "right_handed",
                               [camera_position, look_at, up_vector]):
    camera_position = tf.convert_to_tensor(value=camera_position)
    look_at = tf.convert_to_tensor(value=look_at)
    up_vector = tf.convert_to_tensor(value=up_vector)

    shape.check_static(
        tensor=camera_position,
        tensor_name="camera_position",
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=look_at, tensor_name="look_at", has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=up_vector, tensor_name="up_vector", has_dim_equals=(-1, 3))
    shape.compare_batch_dimensions(
        tensors=(camera_position, look_at, up_vector),
        last_axes=-2,
        tensor_names=("camera_position", "look_at", "up_vector"),
        broadcast_compatible=False)

    z_axis = tf.linalg.l2_normalize(look_at - camera_position, axis=-1)
    horizontal_axis = tf.linalg.l2_normalize(
        vector.cross(z_axis, up_vector), axis=-1)
    vertical_axis = vector.cross(horizontal_axis, z_axis)

    batch_shape = tf.shape(input=horizontal_axis)[:-1]
    zeros = tf.zeros(
        shape=tf.concat((batch_shape, (3,)), axis=-1),
        dtype=horizontal_axis.dtype)
    one = tf.ones(
        shape=tf.concat((batch_shape, (1,)), axis=-1),
        dtype=horizontal_axis.dtype)

    matrix = tf.concat(
        (horizontal_axis, -vector.dot(horizontal_axis, camera_position),
         vertical_axis, -vector.dot(vertical_axis, camera_position), -z_axis,
         vector.dot(z_axis, camera_position), zeros, one),
        axis=-1)
    matrix_shape = tf.shape(input=matrix)
    output_shape = tf.concat((matrix_shape[:-1], (4, 4)), axis=-1)
    return tf.reshape(matrix, shape=output_shape)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for slerp."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.math.interpolation import slerp
from tensorflow_graphics.util import test_case

_SQRT2_DIV2 = np.sqrt(2.0).astype(np.float32) * 0.5


class SlerpTest(test_case.TestCase):

  def _pick_random_quaternion(self):
    """Creates a random quaternion with random shape."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    return np.random.normal(size=tensor_shape + [4])

  def _quaternion_slerp_helper(self, q1, q2, p):
    """Calls interpolate function for quaternions."""
    return slerp.interpolate(q1, q2, p, slerp.InterpolationType.QUATERNION)

  def _vector_slerp_helper(self, q1, q2, p):
    """Calls interpolate function for vectors."""
    return slerp.interpolate(q1, q2, p, slerp.InterpolationType.VECTOR)

  def test_interpolate_raises_exceptions(self):
    """Tests if unknown methods raise exceptions."""
    vector1 = self._pick_random_quaternion()
    self.assert_exception_is_raised(
        slerp.interpolate,
        error_msg="Unknown interpolation type supplied.",
        shapes=[],
        vector1=vector1,
        vector2=-vector1,
        percent=0.1,
        method=2)

  def test_interpolate_with_weights_quaternion_preset(self):
    """Compares interpolate to quaternion_weights + interpolate_with_weights."""
    q1 = self._pick_random_quaternion()
    q2 = q1 + tf.ones_like(q1)
    q1 = tf.nn.l2_normalize(q1, axis=-1)
    q2 = tf.nn.l2_normalize(q2, axis=-1)

    weight1, weight2 = slerp.quaternion_weights(q1, q2, 0.25)
    qf = slerp.interpolate_with_weights(q1, q2, weight1, weight2)
    qi = slerp.interpolate(
        q1, q2, 0.25, method=slerp.InterpolationType.QUATERNION)

    self.assertAllClose(qf, qi, atol=1e-9)

  def test_interpolate_with_weights_vector_preset(self):
    """Compares interpolate to vector_weights + interpolate_with_weights."""
    # Any quaternion is a valid vector
    q1 = self._pick_random_quaternion()
    q2 = q1 + tf.ones_like(q1)

    weight1, weight2 = slerp.vector_weights(q1, q2, 0.75)
    qf = slerp.interpolate_with_weights(q1, q2, weight1, weight2)

    qi = slerp.interpolate(q1, q2, 0.75, method=slerp.InterpolationType.VECTOR)
    self.assertAllClose(qf, qi, atol=1e-9)

  @parameterized.parameters(
      # Orthogonal, same hemisphere
      (((1.0, 0.0, 0.0, 0.0), (0.0, 1.0, 0.0, 0.0), (0.5,)),
       ((_SQRT2_DIV2, _SQRT2_DIV2, 0.0, 0.0),)),
      (((_SQRT2_DIV2, _SQRT2_DIV2, 0.0, 0.0),
        (0.0, 0.0, _SQRT2_DIV2, _SQRT2_DIV2), (0.5,)), ((0.5, 0.5, 0.5, 0.5),)),
      # Same hemisphere
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (0.0, 0.0, _SQRT2_DIV2, _SQRT2_DIV2), (0.5,)),
       ((0.408248290463863, 0.0, 0.816496580927726, 0.408248290463863),)),
      # Same quaternions
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0), (0.75,)),
       ((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),)),
      # Anti-polar - small percent
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (-_SQRT2_DIV2, 0.0, -_SQRT2_DIV2, 0.0), (0.2,)),
       ((-_SQRT2_DIV2, 0.0, -_SQRT2_DIV2, 0.0),)),
      # Anti-polar - large percent
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (-_SQRT2_DIV2, 0.0, -_SQRT2_DIV2, 0.0), (0.8,)),
       ((-_SQRT2_DIV2, 0.0, -_SQRT2_DIV2, 0.0),)),
      # Extrapolation - same hemisphere
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (_SQRT2_DIV2, _SQRT2_DIV2, 0.0, 0.0), (-0.5,)),
       ((0.408248290463863, -0.408248290463863, 0.816496580927726, 0.0),)),
      # Extrapolation - opposite hemisphere
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (-_SQRT2_DIV2, _SQRT2_DIV2, 0.0, 0.0), (-0.5,)),
       ((-0.408248290463863, -0.408248290463863, -0.816496580927726, 0.0),)),
  )
  def test_quaternion_slerp_preset(self, test_inputs, test_outputs):
    """Tests the accuracy of qslerp against numpy-quaternion values."""
    test_inputs = [np.array(test_input).astype(np.float32)
                   for test_input in test_inputs]
    self.assert_output_is_correct(self._quaternion_slerp_helper, test_inputs,
                                  test_outputs, tile=False)

  def test_unnormalized_quaternion_weights_exception_raised(self):
    """Tests if quaternion_weights raise exceptions for unnormalized input."""
    q1 = self._pick_random_quaternion()
    q2 = tf.nn.l2_normalize(q1, axis=-1)
    p = tf.constant((0.5), dtype=q1.dtype)

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(slerp.quaternion_weights(q1, q2, p))

  @parameterized.parameters(
      ((4,), (4,), (1,)),
      ((None, 4), (None, 4), (None, 1)),
      ((None, 4), (None, 4), (None, 4)),
  )
  def test_quaternion_weights_exception_not_raised(self, *shapes):
    """Tests that valid input shapes do not raise exceptions for qslerp."""
    self.assert_exception_is_not_raised(slerp.quaternion_weights, shapes)

  @parameterized.parameters(
      ("must have exactly 4 dimensions in axis -1", (3,), (4,), (1,)),
      ("must have exactly 4 dimensions in axis -1", (4,), (3,), (1,)),
      ("Not all batch dimensions are broadcast-compatible.", (2, 4), (3, 4),
       (1,)),
      ("Not all batch dimensions are broadcast-compatible.", (1, 4), (3, 4),
       (2,)),
  )
  def test_quaternion_weights_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised for qslerp."""
    self.assert_exception_is_raised(slerp.quaternion_weights, error_msg, shapes)

  @parameterized.parameters(
      # Same quaternions
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0), (0.75,)), (
            (0.25,),
            (0.75,),
        )),
      # Anti-polar - small percent
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (-_SQRT2_DIV2, 0.0, -_SQRT2_DIV2, 0.0), (0.2,)), (
            (-0.8,),
            (0.2,),
        )),
      # Anti-polar - large percent
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (-_SQRT2_DIV2, 0.0, -_SQRT2_DIV2, 0.0), (0.8,)), (
            (-0.2,),
            (0.8,),
        )),
  )
  def test_quaternion_weights_preset(self, test_inputs, test_outputs):
    """Tests the accuracy of quaternion_weights for problem cases."""
    test_inputs = [np.array(test_input).astype(np.float32)
                   for test_input in test_inputs]
    self.assert_output_is_correct(slerp.quaternion_weights, test_inputs,
                                  test_outputs, tile=False)

  @parameterized.parameters(
      ((3,), (3,), (1,)),
      ((None, 4), (None, 4), (None, 1)),
  )
  def test_vector_weights_exception_not_raised(self, *shapes):
    """Tests that valid inputs do not raise exceptions for vector_weights."""
    self.assert_exception_is_not_raised(slerp.vector_weights, shapes)

  @parameterized.parameters(
      ("must have the same number of dimensions in axes", (None, 3), (None, 4),
       (1,)),
      ("must have the same number of dimensions in axes", (2, 3), (2, 4), (1,)),
      ("Not all batch dimensions are broadcast-compatible.", (2, 3), (3, 3),
       (1,)),
      ("Not all batch dimensions are broadcast-compatible.", (1, 3), (3, 3),
       (2,)),
  )
  def test_vector_weights_exception_raised(self, error_msg, *shapes):
    """Tests that shape exceptions are properly raised for vector_weights."""
    self.assert_exception_is_raised(slerp.vector_weights, error_msg, shapes)

  @parameterized.parameters(
      # Orthogonal, same hemisphere
      (((1.0, 0.0, 0.0, 0.0), (0.0, 1.0, 0.0, 0.0), (0.5,)),
       ((_SQRT2_DIV2, _SQRT2_DIV2, 0.0, 0.0),)),
      (((_SQRT2_DIV2, _SQRT2_DIV2, 0.0, 0.0),
        (0.0, 0.0, _SQRT2_DIV2, _SQRT2_DIV2), (0.5,)), ((0.5, 0.5, 0.5, 0.5),)),
      # Same hemisphere
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (0.0, 0.0, _SQRT2_DIV2, _SQRT2_DIV2), (0.5,)),
       ((0.408248290463863, 0.0, 0.816496580927726, 0.408248290463863),)),
      # Same vectors
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0), (0.75,)),
       ((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),)),
      # Anti-polar - equal weights
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (-_SQRT2_DIV2, 0.0, -_SQRT2_DIV2, 0.0), (0.5,)),
       ((0.0, 0.0, 0.0, 0.0),)),
      # Anti-polar - small percent
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (-_SQRT2_DIV2, 0.0, -_SQRT2_DIV2, 0.0), (0.25,)),
       ((0.5, 0.0, 0.5, 0.0),)),
      # Extrapolation - same hemisphere
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (_SQRT2_DIV2, _SQRT2_DIV2, 0.0, 0.0), (-1.0,)),
       ((0.0, -_SQRT2_DIV2, _SQRT2_DIV2, 0.0),)),
      # Extrapolation - opposite hemisphere
      (((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0),
        (-_SQRT2_DIV2, _SQRT2_DIV2, 0.0, 0.0), (1.5,)),
       ((-_SQRT2_DIV2, -0.0, -_SQRT2_DIV2, 0.0),)),
      # Unnormalized vectors
      (((4.0, 0.0), (0.0, 1.0), (0.5,)), ((2.82842712, _SQRT2_DIV2),)),
  )
  def test_vector_slerp_preset(self, test_inputs, test_outputs):
    """Tests the accuracy of vector slerp results."""
    test_inputs = [np.array(test_input).astype(np.float32)
                   for test_input in test_inputs]
    self.assert_output_is_correct(self._vector_slerp_helper, test_inputs,
                                  test_outputs, tile=False)

  def test_vector_weights_reduce_to_lerp_preset(self):
    """Tests if vector slerp reduces to lerp for identical vectors as input."""
    q1 = tf.constant((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0))
    q2 = tf.constant((_SQRT2_DIV2, 0.0, _SQRT2_DIV2, 0.0))
    p = tf.constant((0.75,), dtype=q1.dtype)

    w1, w2 = slerp.vector_weights(q1, q2, p)

    self.assertAllClose(w1, (0.25,), rtol=1e-6)
    self.assertAllClose(w2, (0.75,), rtol=1e-6)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the visual hull voxel rendering."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def render(voxels, name=None):
  """Renders the visual hull of a voxel grid, as described in ["Escaping Plato's Cave: 3D Shape From Adversarial Rendering" (Henzler 2019)](https://github.com/henzler/platonicgan).

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    voxels: A tensor of shape `[A1, ..., An, Vx, Vy, Vz, Vd]`, where Vx, Vy, Vz
      are the dimensions of the voxel grid and Vd the dimension of the
      information stored in each voxel (e.g. 3 for RGB color).
    name: A name for this op. Defaults to "visual_hull_render".

  Returns:
    A tensor of shape `[A1, ..., An, Vx, Vy, Vd]` representing images of size
    (Vx,Vy).

  Raises:
    ValueError: If the shape of the input tensors are not supported.
  """
  with tf.compat.v1.name_scope(name, "visual_hull_render", [voxels]):
    voxels = tf.convert_to_tensor(value=voxels)

    shape.check_static(
        tensor=voxels, tensor_name="voxels", has_rank_greater_than=3)

    image = tf.reduce_sum(input_tensor=voxels, axis=-2)
    image = tf.ones_like(image) - tf.math.exp(-image)
    return image


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with test data for transformation tests."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

ANGLE_0 = np.array((0.,))
ANGLE_45 = np.array((np.pi / 4.,))
ANGLE_90 = np.array((np.pi / 2.,))
ANGLE_180 = np.array((np.pi,))

AXIS_2D_0 = np.array((0., 0.))
AXIS_2D_X = np.array((1., 0.))
AXIS_2D_Y = np.array((0., 1.))


def _rotation_2d_x(angle):
  """Creates a 2d rotation matrix.

  Args:
    angle: The angle.

  Returns:
    The 2d rotation matrix.
  """
  angle = angle.item()
  return np.array(((np.cos(angle), -np.sin(angle)),
                   (np.sin(angle), np.cos(angle))))  # pyformat: disable


MAT_2D_ID = np.eye(2)
MAT_2D_45 = _rotation_2d_x(ANGLE_45)
MAT_2D_90 = _rotation_2d_x(ANGLE_90)
MAT_2D_180 = _rotation_2d_x(ANGLE_180)

AXIS_3D_0 = np.array((0., 0., 0.))
AXIS_3D_X = np.array((1., 0., 0.))
AXIS_3D_Y = np.array((0., 1., 0.))
AXIS_3D_Z = np.array((0., 0., 1.))


def _axis_angle_to_quaternion(axis, angle):
  """Converts an axis-angle representation to a quaternion.

  Args:
    axis: The axis of rotation.
    angle: The angle.

  Returns:
    The quaternion.
  """
  quat = np.zeros(4)
  quat[0:3] = axis * np.sin(0.5 * angle)
  quat[3] = np.cos(0.5 * angle)
  return quat


QUAT_ID = _axis_angle_to_quaternion(AXIS_3D_0, ANGLE_0)
QUAT_X_45 = _axis_angle_to_quaternion(AXIS_3D_X, ANGLE_45)
QUAT_X_90 = _axis_angle_to_quaternion(AXIS_3D_X, ANGLE_90)
QUAT_X_180 = _axis_angle_to_quaternion(AXIS_3D_X, ANGLE_180)
QUAT_Y_45 = _axis_angle_to_quaternion(AXIS_3D_Y, ANGLE_45)
QUAT_Y_90 = _axis_angle_to_quaternion(AXIS_3D_Y, ANGLE_90)
QUAT_Y_180 = _axis_angle_to_quaternion(AXIS_3D_Y, ANGLE_180)
QUAT_Z_45 = _axis_angle_to_quaternion(AXIS_3D_Z, ANGLE_45)
QUAT_Z_90 = _axis_angle_to_quaternion(AXIS_3D_Z, ANGLE_90)
QUAT_Z_180 = _axis_angle_to_quaternion(AXIS_3D_Z, ANGLE_180)


def _rotation_3d_x(angle):
  """Creates a 3d rotation matrix around the x axis.

  Args:
    angle: The angle.

  Returns:
    The 3d rotation matrix.
  """
  angle = angle.item()
  return np.array(((1., 0., 0.),
                   (0., np.cos(angle), -np.sin(angle)),
                   (0., np.sin(angle), np.cos(angle))))  # pyformat: disable


def _rotation_3d_y(angle):
  """Creates a 3d rotation matrix around the y axis.

  Args:
    angle: The angle.

  Returns:
    The 3d rotation matrix.
  """
  angle = angle.item()
  return np.array(((np.cos(angle), 0., np.sin(angle)),
                   (0., 1., 0.),
                   (-np.sin(angle), 0., np.cos(angle))))  # pyformat: disable


def _rotation_3d_z(angle):
  """Creates a 3d rotation matrix around the z axis.

  Args:
    angle: The angle.

  Returns:
    The 3d rotation matrix.
  """
  angle = angle.item()
  return np.array(((np.cos(angle), -np.sin(angle), 0.),
                   (np.sin(angle), np.cos(angle), 0.),
                   (0., 0., 1.)))  # pyformat: disable


MAT_3D_ID = np.eye(3)
MAT_3D_X_45 = _rotation_3d_x(ANGLE_45)
MAT_3D_X_90 = _rotation_3d_x(ANGLE_90)
MAT_3D_X_180 = _rotation_3d_x(ANGLE_180)
MAT_3D_Y_45 = _rotation_3d_y(ANGLE_45)
MAT_3D_Y_90 = _rotation_3d_y(ANGLE_90)
MAT_3D_Y_180 = _rotation_3d_y(ANGLE_180)
MAT_3D_Z_45 = _rotation_3d_z(ANGLE_45)
MAT_3D_Z_90 = _rotation_3d_z(ANGLE_90)
MAT_3D_Z_180 = _rotation_3d_z(ANGLE_180)



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""This modules implements Euler angles functionalities.

The Euler angles are defined using a vector $$[\theta, \gamma, \beta]^T \in
\mathbb{R}^3$$, where $$\theta$$ is the angle about $$x$$, $$\gamma$$ the angle
about $$y$$, and $$\beta$$ is the angle about $$z$$

More details about Euler angles can be found on [this page.]
(https://en.wikipedia.org/wiki/Euler_angles)

Note: The angles are defined in radians.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math

import tensorflow as tf

from tensorflow_graphics.geometry.transformation import quaternion
from tensorflow_graphics.geometry.transformation import rotation_matrix_3d
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import safe_ops
from tensorflow_graphics.util import shape


def from_axis_angle(axis, angle, name=None):
  """Converts axis-angle to Euler angles.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    axis: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents a normalized axis.
    angle: A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      represents an angle.
    name: A name for this op that defaults to "euler_from_axis_angle".

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents
    the three Euler angles.
  """
  with tf.compat.v1.name_scope(name, "euler_from_axis_angle", [axis, angle]):
    return from_quaternion(quaternion.from_axis_angle(axis, angle))


def from_quaternion(quaternions, name=None):
  """Converts quaternions to Euler angles.

  Args:
    quaternions: A tensor of shape `[A1, ..., An, 4]`, where the last dimension
      represents a normalized quaternion.
    name: A name for this op that defaults to "euler_from_quaternion".

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents
    the three Euler angles.
  """

  def general_case(r00, r10, r21, r22, r20, eps_addition):
    """Handles the general case."""
    theta_y = -tf.asin(r20)
    sign_cos_theta_y = safe_ops.nonzero_sign(tf.cos(theta_y))
    r00 = safe_ops.nonzero_sign(r00) * eps_addition + r00
    r22 = safe_ops.nonzero_sign(r22) * eps_addition + r22
    theta_z = tf.atan2(r10 * sign_cos_theta_y, r00 * sign_cos_theta_y)
    theta_x = tf.atan2(r21 * sign_cos_theta_y, r22 * sign_cos_theta_y)
    return tf.stack((theta_x, theta_y, theta_z), axis=-1)

  def gimbal_lock(r01, r02, r20, eps_addition):
    """Handles Gimbal locks."""
    sign_r20 = safe_ops.nonzero_sign(r20)
    r02 = safe_ops.nonzero_sign(r02) * eps_addition + r02
    theta_x = tf.atan2(-sign_r20 * r01, -sign_r20 * r02)
    theta_y = -sign_r20 * tf.constant(math.pi / 2.0, dtype=r20.dtype)
    theta_z = tf.zeros_like(theta_x)
    angles = tf.stack((theta_x, theta_y, theta_z), axis=-1)
    return angles

  with tf.compat.v1.name_scope(name, "euler_from_quaternion", [quaternions]):
    quaternions = tf.convert_to_tensor(value=quaternions)

    shape.check_static(
        tensor=quaternions,
        tensor_name="quaternions",
        has_dim_equals=(-1, 4))

    x, y, z, w = tf.unstack(quaternions, axis=-1)
    tx = safe_ops.safe_shrink(2.0 * x, -2.0, 2.0, True)
    ty = safe_ops.safe_shrink(2.0 * y, -2.0, 2.0, True)
    tz = safe_ops.safe_shrink(2.0 * z, -2.0, 2.0, True)
    twx = tx * w
    twy = ty * w
    twz = tz * w
    txx = tx * x
    txy = ty * x
    txz = tz * x
    tyy = ty * y
    tyz = tz * y
    tzz = tz * z

    # The following is clipped due to numerical instabilities that can take some
    # enties outside the [-1;1] range.
    r00 = safe_ops.safe_shrink(1.0 - (tyy + tzz), -1.0, 1.0, True)
    r10 = safe_ops.safe_shrink(txy + twz, -1.0, 1.0, True)
    r21 = safe_ops.safe_shrink(tyz + twx, -1.0, 1.0, True)
    r22 = safe_ops.safe_shrink(1.0 - (txx + tyy), -1.0, 1.0, True)
    r20 = safe_ops.safe_shrink(txz - twy, -1.0, 1.0, True)
    r01 = safe_ops.safe_shrink(txy - twz, -1.0, 1.0, True)
    r02 = safe_ops.safe_shrink(txz + twy, -1.0, 1.0, True)
    eps_addition = asserts.select_eps_for_addition(quaternions.dtype)
    general_solution = general_case(r00, r10, r21, r22, r20, eps_addition)
    gimbal_solution = gimbal_lock(r01, r02, r20, eps_addition)
    # The general solution is unstable close to the Gimbal lock, and the gimbal
    # solution is not toooff in these cases.
    is_gimbal = tf.less(tf.abs(tf.abs(r20) - 1.0), 1.0e-6)
    gimbal_mask = tf.stack((is_gimbal, is_gimbal, is_gimbal), axis=-1)
    return tf.compat.v1.where(gimbal_mask, gimbal_solution, general_solution)


def from_rotation_matrix(rotation_matrix, name=None):
  """Converts rotation matrices to Euler angles.

  The rotation matrices are assumed to have been constructed by rotation around
  the $$x$$, then $$y$$, and finally the $$z$$ axis.

  Note:
    There is an infinite number of solutions to this problem. There are
  Gimbal locks when abs(rotation_matrix(2,0)) == 1, which are not handled.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    rotation_matrix: A tensor of shape `[A1, ..., An, 3, 3]`, where the last two
      dimensions represent a rotation matrix.
    name: A name for this op that defaults to "euler_from_rotation_matrix".

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents
    the three Euler angles.

  Raises:
    ValueError: If the shape of `rotation_matrix` is not supported.
  """

  def general_case(rotation_matrix, r20, eps_addition):
    """Handles the general case."""
    theta_y = -tf.asin(r20)
    sign_cos_theta_y = safe_ops.nonzero_sign(tf.cos(theta_y))
    r00 = rotation_matrix[..., 0, 0]
    r10 = rotation_matrix[..., 1, 0]
    r21 = rotation_matrix[..., 2, 1]
    r22 = rotation_matrix[..., 2, 2]
    r00 = safe_ops.nonzero_sign(r00) * eps_addition + r00
    r22 = safe_ops.nonzero_sign(r22) * eps_addition + r22
    # cos_theta_y evaluates to 0 on Gimbal locks, in which case the output of
    # this function will not be used.
    theta_z = tf.atan2(r10 * sign_cos_theta_y, r00 * sign_cos_theta_y)
    theta_x = tf.atan2(r21 * sign_cos_theta_y, r22 * sign_cos_theta_y)
    angles = tf.stack((theta_x, theta_y, theta_z), axis=-1)
    return angles

  def gimbal_lock(rotation_matrix, r20, eps_addition):
    """Handles Gimbal locks."""
    r01 = rotation_matrix[..., 0, 1]
    r02 = rotation_matrix[..., 0, 2]
    sign_r20 = safe_ops.nonzero_sign(r20)
    r02 = safe_ops.nonzero_sign(r02) * eps_addition + r02
    theta_x = tf.atan2(-sign_r20 * r01, -sign_r20 * r02)
    theta_y = -sign_r20 * tf.constant(math.pi / 2.0, dtype=r20.dtype)
    theta_z = tf.zeros_like(theta_x)
    angles = tf.stack((theta_x, theta_y, theta_z), axis=-1)
    return angles

  with tf.compat.v1.name_scope(name, "euler_from_rotation_matrix",
                               [rotation_matrix]):
    rotation_matrix = tf.convert_to_tensor(value=rotation_matrix)

    shape.check_static(
        tensor=rotation_matrix,
        tensor_name="rotation_matrix",
        has_rank_greater_than=1,
        has_dim_equals=((-1, 3), (-2, 3)))
    rotation_matrix = rotation_matrix_3d.assert_rotation_matrix_normalized(
        rotation_matrix)

    r20 = rotation_matrix[..., 2, 0]
    eps_addition = asserts.select_eps_for_addition(rotation_matrix.dtype)
    general_solution = general_case(rotation_matrix, r20, eps_addition)
    gimbal_solution = gimbal_lock(rotation_matrix, r20, eps_addition)
    is_gimbal = tf.equal(tf.abs(r20), 1)
    gimbal_mask = tf.stack((is_gimbal, is_gimbal, is_gimbal), axis=-1)
    return tf.compat.v1.where(gimbal_mask, gimbal_solution, general_solution)


def inverse(euler_angle, name=None):
  """Computes the angles that would inverse a transformation by euler_angle.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    euler_angle: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the three Euler angles.
    name: A name for this op that defaults to "euler_inverse".

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents
    the three Euler angles.

  Raises:
    ValueError: If the shape of `euler_angle` is not supported.
  """
  with tf.compat.v1.name_scope(name, "euler_inverse", [euler_angle]):
    euler_angle = tf.convert_to_tensor(value=euler_angle)

    shape.check_static(
        tensor=euler_angle,
        tensor_name="euler_angle",
        has_dim_equals=(-1, 3))

    return -euler_angle


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the recall metric."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import numpy as np

from tensorflow_graphics.nn.metric import recall
from tensorflow_graphics.util import test_case


def random_tensor(tensor_shape):
  return np.random.uniform(low=0.0, high=1.0, size=tensor_shape)


def random_tensor_shape():
  tensor_size = np.random.randint(5) + 1
  return np.random.randint(1, 10, size=(tensor_size)).tolist()


class RecallTest(test_case.TestCase):

  @parameterized.parameters(
      # recall = 0.25.
      ((0, 1, 1, 1, 1), (1, 1, 0, 0, 0), 0.25),
      # recall = 1.
      ((0, 0, 0, 1, 1, 1, 0, 1), (0, 0, 0, 1, 1, 1, 0, 1), 1),
      # All-0 predictions, returns 0.
      ((0, 1, 0, 0, 0, 0), (0, 0, 0, 0, 0, 0), 0),
      # All-0 ground truth, returns 0.
      ((0, 0, 0, 0, 0, 0), (0, 0, 0, 1, 0, 1), 0),
  )
  def test_evaluate_preset(self, ground_truth, predictions, expected_recall):
    tensor_shape = random_tensor_shape()

    ground_truth_labels = np.tile(ground_truth, tensor_shape + [1])
    predicted_labels = np.tile(predictions, tensor_shape + [1])
    expected = np.tile(expected_recall, tensor_shape)

    result = recall.evaluate(ground_truth_labels, predicted_labels, classes=[1])

    self.assertAllClose(expected, result)

  @parameterized.parameters(
      # Recall for classes 2, 3: [1/3, 1.]
      ((2, 0, 3, 1, 2, 2), (2, 3, 3, 2, 3, 3), [2, 3], False, [1. / 3, 1]),
      # Average recall for classes 2, 3: 2/3
      ((2, 0, 3, 1, 2, 2), (2, 3, 3, 2, 3, 3), [2, 3], True, 2. / 3),
      # Recall for all classes: [0, 0, 0.5, 0.5]
      ((1, 2, 3, 3, 1, 1, 2),
       (0, 2, 0, 3, 0, 2, 0), None, False, [0, 0, 0.5, 0.5]),
      # Average recall for all classes: 0.25
      ((1, 2, 3, 3, 1, 1, 2), (0, 2, 0, 3, 0, 2, 0), None, True, 0.25),
  )
  def test_evaluate_preset_multiclass(self, ground_truth, predictions, classes,
                                      reduce_average, expected_recall):
    tensor_shape = random_tensor_shape()

    ground_truth_labels = np.tile(ground_truth, tensor_shape + [1])
    predicted_labels = np.tile(predictions, tensor_shape + [1])
    expected = np.tile(expected_recall,
                       tensor_shape + ([1] if not reduce_average else []))

    result = recall.evaluate(ground_truth_labels, predicted_labels, classes,
                             reduce_average)

    self.assertAllClose(expected, result)

  @parameterized.parameters(
      ("Not all batch dimensions are broadcast-compatible.", (1, 5, 3), (4, 3)),
      ("Not all batch dimensions are broadcast-compatible.", (3, 4), (2, 4, 5)),
  )
  def test_evaluate_shape_exception_raised(self, error_msg, *shape):
    """Tests that the shape exception is raised."""
    self.assert_exception_is_raised(recall.evaluate, error_msg, shape)

  @parameterized.parameters(
      ((1, 5, 3), (2, 5, 1)),
      ((None, 2, 6), (4, 2, None)),
      ((3, 1, 1, 2), (3, 5, 8, 2)),
  )
  def test_evaluate_shape_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(recall.evaluate, shapes)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A collection of training helper utilities."""

from __future__ import print_function

import argparse
import os
import tempfile
import time

import tensorflow as tf
import termcolor


class ArgumentParser(argparse.ArgumentParser):
  """Argument parser with default flags, and tensorboard helpers."""

  def __init__(self, *args, **kwargs):
    argparse.ArgumentParser.__init__(self, *args, **kwargs)

    # --- Query default logdir
    random_logdir = tempfile.mkdtemp(prefix="tensorboard_")
    default_logdir = os.environ.get("TENSORBOARD_DEFAULT_LOGDIR", random_logdir)

    # --- Add the default options
    self.add("--logdir", default_logdir, help="tensorboard dir")
    self.add("--tensorboard", True, help="should generate summaries?")
    self.add("--assert_gpu", True, help="asserts on missing GPU accelerator")
    self.add("--tf_quiet", True, help="no verbose tf startup")

  def add(self, name, default, **kwargs):
    """More compact argumentparser 'add' flag method."""
    helpstring = kwargs["help"] if "help" in kwargs else ""
    metavar = kwargs["metavar"] if "metavar" in kwargs else name

    # --- Fixes problems with bool arguments
    def str2bool(string):
      if isinstance(string, bool):
        return str
      if string.lower() in ("true", "yes"):
        return True
      if string.lower() in ("false", "no"):
        return False
      raise argparse.ArgumentTypeError("Bad value for boolean flag")

    mytype = type(default)
    if isinstance(default, bool):
      mytype = str2bool

    self.add_argument(
        name, metavar=metavar, default=default, help=helpstring, type=mytype)

  def parse_args(self, args=None, namespace=None):
    """WARNING: programmatically changes the logdir flags."""
    flags = super(ArgumentParser, self).parse_args(args)

    # --- setup automatic logdir (timestamp)
    if "timestamp" in flags.logdir:
      timestamp = time.strftime("%a%d_%H:%M:%S")  # "Tue19_12:02:26"
      flags.logdir = flags.logdir.replace("timestamp", timestamp)

    if flags.tf_quiet:
      set_tensorflow_log_level(3)

    if flags.assert_gpu:
      assert_gpu_available()

    # --- ensure logdir ends in /
    if flags.logdir[-1] != "/":
      flags.logdir += "/"

    return flags


def assert_gpu_available():
  """Verifies a GPU accelerator is available."""
  physical_devices = tf.config.list_physical_devices("GPU")
  num_gpus = len(physical_devices)
  assert num_gpus >= 1, "execution requires one GPU"


def set_tensorflow_log_level(level=3):
  """Sets the log level of TensorFlow."""
  os.environ["TF_CPP_MIN_LOG_LEVEL"] = str(level)


def summary_command(parser, flags, log_to_file=True, log_to_summary=True):
  """Cache the command used to reproduce experiment in summary folder."""
  if not flags.tensorboard:
    return
  exec_string = "python " + parser.prog + " \\\n"
  nflags = len(vars(flags))
  for i, arg in enumerate(vars(flags)):
    exec_string += "  --{} ".format(arg)
    exec_string += "{}".format(getattr(flags, arg))
    if i + 1 < nflags:
      exec_string += " \\\n"
  exec_string += "\n"
  if log_to_file:
    with tf.io.gfile.GFile(
        os.path.join(flags.logdir, "command.txt"), mode="w") as fid:
      fid.write(exec_string)
  if log_to_summary and flags.tensorboard:
    tf.summary.text("command", exec_string, step=0)


def setup_tensorboard(flags):
  """Creates summary writers, and setups default tensorboard paths."""
  if not flags.tensorboard:
    return

  # --- Do not allow experiment with same name
  assert (not tf.io.gfile.exists(flags.logdir) or
          not tf.io.gfile.listdir(flags.logdir)), \
    "CRITICAL: folder {} already exists".format(flags.logdir)

  # --- Log where summary can be found
  print("View results with: ")
  termcolor.cprint("  tensorboard --logdir {}".format(flags.logdir), "red")
  writer = tf.summary.create_file_writer(flags.logdir, flush_millis=10000)
  writer.set_as_default()

  # --- Log dir name tweak for "hypertune"
  log_dir = ""
  trial_id = int(os.environ.get("CLOUD_ML_TRIAL_ID", 0))
  if trial_id != 0:
    if log_dir.endswith(os.sep):
      log_dir = log_dir[:-1]  # removes trailing "/"
    log_dir += "_trial{0:03d}/".format(trial_id)


def handle_keyboard_interrupt(flags):
  """Informs user how to delete stale summaries."""
  print("Keyboard interrupt by user")
  if flags.logdir.startswith("gs://"):
    bucketpath = flags.logdir[5:]
    print("Delete these summaries with: ")
    termcolor.cprint("  gsutil rm -rf {}".format(flags.logdir), "red")
    baseurl = "  https://pantheon.google.com/storage/browser/{}"
    print("Or by visiting: ")
    termcolor.cprint(baseurl.format(bucketpath), "red")
  else:
    print("Delete these summaries with: ")
    termcolor.cprint("  rm -rf {}".format(flags.logdir), "red")



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for graph convolution ops."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

import tensorflow_graphics.geometry.convolution.graph_convolution as gc
from tensorflow_graphics.util import test_case


def _dense_to_sparse(data):
  """Convert a numpy array to a tf.SparseTensor."""
  indices = np.where(data)
  return tf.SparseTensor(
      np.stack(indices, axis=-1), data[indices], dense_shape=data.shape)


def _dummy_data(batch_size, num_vertices, num_channels):
  """Create inputs for feature_steered_convolution."""
  if batch_size > 0:
    data = np.zeros(
        shape=(batch_size, num_vertices, num_channels), dtype=np.float32)
    neighbors = _dense_to_sparse(
        np.tile(np.eye(num_vertices, dtype=np.float32), (batch_size, 1, 1)))
  else:
    data = np.zeros(shape=(num_vertices, num_channels), dtype=np.float32)
    neighbors = _dense_to_sparse(np.eye(num_vertices, dtype=np.float32))
  return data, neighbors


def _dummy_variables(in_channels, out_channels, num_weight_matrices):
  """Create variable substitutes for feature_steered_convolution."""
  var_u = tf.zeros(shape=(in_channels, num_weight_matrices))
  var_v = tf.zeros(shape=(in_channels, num_weight_matrices))
  var_c = tf.zeros(shape=(num_weight_matrices))
  var_w = tf.zeros(shape=(num_weight_matrices, in_channels, out_channels))
  var_b = tf.zeros(shape=(out_channels))
  return var_u, var_v, var_c, var_w, var_b


def _random_data(batch_size,
                 num_vertices,
                 num_channels,
                 padding,
                 only_self_edges,
                 data_type=np.float32,
                 neighbors_type=np.float32,
                 sizes_type=np.int32):
  """Create random inputs for feature_steered_convolution."""

  def _random_data_2d(padding):
    size = num_vertices if not padding else np.random.randint(
        low=1, high=num_vertices + 1)
    data = np.random.uniform(size=(size, num_channels)).astype(data_type)
    if only_self_edges:
      neighbors = np.eye(size, dtype=neighbors_type)
    else:
      random = np.random.uniform(size=(size, size)).astype(neighbors_type)
      neighbors = np.maximum(
          np.where(random > 0.75, np.ones_like(random), np.zeros_like(random)),
          np.eye(size, dtype=neighbors_type))
      neighbors = neighbors / np.sum(neighbors, axis=1, keepdims=True)
    if padding:
      data = np.pad(data, ((0, num_vertices - size), (0, 0)), "constant")
      neighbors = np.pad(neighbors,
                         ((0, num_vertices - size), (0, num_vertices - size)),
                         "constant")
      return data, neighbors, size
    else:
      return data, neighbors

  if batch_size > 0:
    list_2d = [_random_data_2d(padding=padding) for _ in range(batch_size)]
    data = np.stack([i[0] for i in list_2d], 0).astype(data_type)
    neighbors = np.stack([i[1] for i in list_2d], 0).astype(neighbors_type)
    if padding:
      sizes = np.stack([i[2] for i in list_2d], 0).astype(sizes_type)
      return data, _dense_to_sparse(neighbors), sizes
    else:
      return data, _dense_to_sparse(neighbors)
  else:
    if padding:
      raise ValueError("Padding only allowed with batched data.")
    data, neighbors = _random_data_2d(padding=False)
    return data.astype(data_type), _dense_to_sparse(
        neighbors.astype(neighbors_type))


def _random_variables(in_channels,
                      out_channels,
                      num_weight_matrices,
                      dtype=np.float32):
  """Create random variables for feature_steered_convolution."""

  def _random_constant(shape, dtype):
    return tf.constant(np.random.uniform(size=shape).astype(dtype))

  var_u = _random_constant([in_channels, num_weight_matrices], dtype)
  var_v = _random_constant([in_channels, num_weight_matrices], dtype)
  var_c = _random_constant([num_weight_matrices], dtype)
  var_w = _random_constant([num_weight_matrices, in_channels, out_channels],
                           dtype)
  var_b = _random_constant([out_channels], dtype)
  return var_u, var_v, var_c, var_w, var_b


class GraphConvolutionTestFeatureSteeredConvolutionTests(test_case.TestCase):

  @parameterized.parameters(
      ("'sizes' must have an integer type.", np.float32, np.float32, np.float32,
       np.float32),
      ("'data' must have a float type.", np.int32, np.float32, np.int32,
       np.float32),
      ("'neighbors' and 'data' must have the same type.", np.float32,
       np.float64, np.int32, np.float32),
  )
  def test_feature_steered_convolution_exception_raised_types(
      self, err_msg, data_type, neighbors_type, sizes_type, var_type):
    """Check the type errors for invalid input types."""
    data, neighbors, sizes = _random_data(1, 5, 3, True, False, data_type,
                                          neighbors_type, sizes_type)
    u, v, c, w, b = _random_variables(3, 3, 1, var_type)
    with self.assertRaisesRegexp(TypeError, err_msg):
      _ = gc.feature_steered_convolution(
          data=data,
          neighbors=neighbors,
          sizes=sizes,
          var_u=u,
          var_v=v,
          var_c=c,
          var_w=w,
          var_b=b)

  @parameterized.parameters(
      (np.float32, np.float32, np.int32, np.float32),
      (np.float64, np.float64, np.int32, np.float64),
      (np.float32, np.float32, np.int64, np.float32),
      (np.float64, np.float64, np.int64, np.float64),
  )
  def test_feature_steered_convolution_exception_not_raised_types(
      self, data_type, neighbors_type, sizes_type, var_type):
    """Check there are no exceptions for valid input types."""
    data, neighbors, sizes = _random_data(1, 5, 3, True, False, data_type,
                                          neighbors_type, sizes_type)
    u, v, c, w, b = _random_variables(3, 3, 1, var_type)
    try:
      gc.feature_steered_convolution(
          data=data,
          neighbors=neighbors,
          sizes=sizes,
          var_u=u,
          var_v=v,
          var_c=c,
          var_w=w,
          var_b=b)
    except Exception as e:  # pylint: disable=broad-except
      self.fail("Exception raised: %s" % str(e))

  def test_feature_steered_convolution_exception_raised_shapes(self):
    """Check that invalid input shapes trigger the right exceptions."""
    with self.assertRaisesRegexp(ValueError, "must have a rank of 2"):
      data, neighbors = _dummy_data(1, 5, 2)
      u, v, c, w, b = _dummy_variables(2, 2, 1)
      data = data[0, :]
      _ = gc.feature_steered_convolution(
          data=data,
          neighbors=neighbors,
          sizes=None,
          var_u=u,
          var_v=v,
          var_c=c,
          var_w=w,
          var_b=b)

    with self.assertRaisesRegexp(ValueError, "must have a rank greater than 1"):
      u, v, c, w, b = _dummy_variables(2, 2, 1)
      data = np.ones(shape=(5), dtype=np.float32)
      neighbors = _dense_to_sparse(np.ones(shape=(5), dtype=np.float32))
      _ = gc.feature_steered_convolution(
          data=data,
          neighbors=neighbors,
          sizes=None,
          var_u=u,
          var_v=v,
          var_c=c,
          var_w=w,
          var_b=b)

    with self.assertRaisesRegexp(ValueError,
                                 "Not all batch dimensions are identical."):
      data, neighbors = _dummy_data(1, 5, 2)
      u, v, c, w, b = _dummy_variables(2, 2, 1)
      _ = gc.feature_steered_convolution(
          data=data,
          neighbors=neighbors,
          sizes=(1, 1),
          var_u=u,
          var_v=v,
          var_c=c,
          var_w=w,
          var_b=b)

  @parameterized.parameters(
      (1, 1, 1, 1, 1),
      (4, 2, 3, 6, 5),
      (0, 1, 1, 1, 1),
      (0, 2, 3, 6, 5),
  )
  def test_feature_steered_convolution_output_shape(self, batch_size,
                                                    num_vertices, in_channels,
                                                    out_channels,
                                                    num_weight_matrices):
    """Check that the output of convolution has the correct shape."""
    data, neighbors = _dummy_data(batch_size, num_vertices, in_channels)
    u, v, c, w, b = _dummy_variables(in_channels, out_channels,
                                     num_weight_matrices)

    y = gc.feature_steered_convolution(
        data=data,
        neighbors=neighbors,
        sizes=None,
        var_u=u,
        var_v=v,
        var_c=c,
        var_w=w,
        var_b=b)
    y_shape = y.shape.as_list()

    self.assertEqual(y_shape[-1], out_channels)
    self.assertAllEqual(y_shape[:-1], data.shape[:-1])

  @parameterized.parameters(
      (1, 1, 1, 1, 1),
      (4, 2, 3, 6, 5),
      (0, 1, 1, 1, 1),
      (0, 2, 3, 6, 5),
  )
  def test_feature_steered_convolution_only_self_edges(self, batch_size,
                                                       num_vertices,
                                                       in_channels,
                                                       out_channels,
                                                       num_weight_matrices):
    """Test convolution when the graph only has self edges."""
    data, neighbors = _random_data(
        batch_size,
        num_vertices,
        in_channels,
        padding=False,
        only_self_edges=True)
    u, v, c, w, b = _random_variables(in_channels, out_channels,
                                      num_weight_matrices)

    with self.subTest(name="w=0_expect_output=b"):
      y = gc.feature_steered_convolution(
          data=data,
          neighbors=neighbors,
          sizes=None,
          var_u=u,
          var_v=v,
          var_c=c,
          var_w=tf.zeros_like(w),
          var_b=b)
      y_expected = tf.broadcast_to(b, y.shape)

      self.assertAllEqual(y, y_expected)

    with self.subTest(name="translation_invariant_self_edges"):
      y = gc.feature_steered_convolution(
          data=data,
          neighbors=neighbors,
          sizes=None,
          var_u=u,
          var_v=-u,
          var_c=c,
          var_w=w,
          var_b=b)
      q = tf.reshape(
          tf.exp(c) / tf.reduce_sum(input_tensor=tf.exp(c)),
          (num_weight_matrices, 1, 1))
      if batch_size > 0:
        q_times_w = tf.reduce_sum(input_tensor=q * w, axis=0, keepdims=True)
        q_times_w = tf.tile(q_times_w, (batch_size, 1, 1))
      else:
        q_times_w = tf.reduce_sum(input_tensor=q * w, axis=0)
      y_expected = tf.matmul(data, q_times_w) + tf.broadcast_to(b, y.shape)

      self.assertAllClose(y, y_expected)

    with self.subTest(name="constant_signal"):
      if batch_size > 0:
        constant_data = np.tile(
            np.random.uniform(size=(batch_size, 1,
                                    in_channels)).astype(np.float32),
            (1, num_vertices, 1))
      else:
        constant_data = np.tile(
            np.random.uniform(size=(1, in_channels)).astype(np.float32),
            (num_vertices, 1))
      y = gc.feature_steered_convolution(
          data=constant_data,
          neighbors=neighbors,
          sizes=None,
          var_u=u,
          var_v=v,
          var_c=c,
          var_w=w,
          var_b=b)
      if batch_size > 0:
        y_expected = tf.tile(y[:, :1, :], (1, num_vertices, 1))
      else:
        y_expected = tf.tile(y[:1, :], (num_vertices, 1))

      self.assertAllClose(y, y_expected)

  @parameterized.parameters(
      (((1.0,), (2.0,), (3.0,)), np.ones(shape=(3, 3)) / 3.0, ((0.5,),),
       ((1.3,),), (-0.7,), (((0.8,),),), (3.0,), ((4.6,), (4.6,), (4.6,))),
      (((1.0,), (2.0,), (3.0,)), np.ones(shape=(3, 3)) / 3.0, ((0.5, 0.2),),
       ((0.3, 0.4),), (-0.7, 0.15), (((0.8,),), ((1.1,),)), (3.0,),
       ((5.011706928844621,), (4.971030281984818,), (4.927388658982911,))),
  )
  def test_feature_steered_convolution_padding_preset(self, data, neighbors, u,
                                                      v, c, w, b, expected):
    """Test expected result for preset data and filter values."""
    array = (np.array(i) for i in (data, neighbors, expected))
    data, neighbors, expected = array
    tensors = (tf.convert_to_tensor(value=np.array(i).astype(data.dtype)) \
               for i in (u, v, c, w, b))
    u, v, c, w, b = tensors
    y = gc.feature_steered_convolution(
        data=data,
        neighbors=_dense_to_sparse(neighbors),
        sizes=None,
        var_u=u,
        var_v=v,
        var_c=c,
        var_w=w,
        var_b=b)
    self.assertAllClose(y, expected)

  @parameterized.parameters(
      (1, 5, 1, 1, 1),
      (2, 6, 3, 6, 5),
      (5, 15, 6, 12, 8),
  )
  def test_feature_steered_convolution_padding_random(self, batch_size,
                                                      num_vertices, in_channels,
                                                      out_channels,
                                                      num_weight_matrices):
    """Test mixed topology batches (random vertices and neighbors)."""
    data, neighbors, sizes = _random_data(
        batch_size,
        num_vertices,
        in_channels,
        padding=True,
        only_self_edges=False)
    u, v, c, w, b = _random_variables(in_channels, out_channels,
                                      num_weight_matrices)

    with self.subTest(name="if_w_is_0_then_y_is_b"):
      y = gc.feature_steered_convolution(
          data=data,
          neighbors=neighbors,
          sizes=sizes,
          var_u=u,
          var_v=v,
          var_c=c,
          var_w=tf.zeros_like(w),
          var_b=b)
      for k in range(batch_size):
        y_crop = y[k, :sizes[k], :]
        y_expected = tf.broadcast_to(b, y_crop.shape)

        self.assertAllEqual(y_crop, y_expected)
        # Check for zeros in the padded region.
        self.assertAllEqual(y[k, sizes[k]:, :],
                            tf.zeros((num_vertices - sizes[k], out_channels)))

    with self.subTest(name="convolve_with_constant"):
      constant_data = data
      for k in range(batch_size):
        constant_data[k, :sizes[k], :] = np.tile(data[k, 0, :], (sizes[k], 1))

      y = gc.feature_steered_convolution(
          data=constant_data,
          neighbors=neighbors,
          sizes=sizes,
          var_u=u,
          var_v=v,
          var_c=c,
          var_w=w,
          var_b=b)
      for k in range(batch_size):
        y_crop = y[k, :sizes[k], :]
        y_const = tf.broadcast_to(y_crop[0, :], y_crop.shape)

        self.assertAllClose(y_crop, y_const)
        # Check for zeros in the padded region.
        self.assertAllEqual(y[k, sizes[k]:, :],
                            tf.zeros([num_vertices - sizes[k], out_channels]))

  @parameterized.parameters(
      (1, 10, 3, 1, True),
      (3, 6, 1, 4, True),
      (0, 10, 5, 2, False),
      (1, 10, 3, 1, False),
      (3, 6, 1, 4, False),
      (0, 10, 5, 2, False),
  )
  def test_feature_steered_convolution_jacobian_random(self, batch_size,
                                                       num_vertices,
                                                       in_channels,
                                                       num_weight_matrices,
                                                       padding):
    """Test the jacobian for random input data."""
    random_data = _random_data(
        batch_size,
        num_vertices,
        in_channels,
        padding,
        only_self_edges=False,
        data_type=np.float64,
        neighbors_type=np.float64)
    data_init = random_data[0]
    neighbors = random_data[1]
    sizes = None if not padding else random_data[2]
    u, v, c, w, b = _random_variables(
        in_channels, in_channels, num_weight_matrices, dtype=np.float64)

    def feature_steered_convolution(data):
      return gc.feature_steered_convolution(
          data=data,
          neighbors=neighbors,
          sizes=sizes,
          var_u=u,
          var_v=v,
          var_c=c,
          var_w=w,
          var_b=b)

    self.assert_jacobian_is_correct_fn(feature_steered_convolution, [data_init])

  @parameterized.parameters(
      (1, 1, 0.0),
      (5, 1, 0.0),
      (1, 3, 0.0),
      (5, 3, 0.0),
      (1, 1, 1.0),
      (5, 1, 1.0),
      (1, 3, 1.0),
      (5, 3, 1.0),
  )
  def test_feature_steered_convolution_jacobian_preset(self, num_vertices,
                                                       num_channels,
                                                       data_multiplier):
    """Test the jacobian is correct for preset inputs."""
    # Corner cases include one vertex, one channel, and all-zero features.
    data_init = data_multiplier * np.random.uniform(
        size=(num_vertices, num_channels)).astype(np.float64)
    neighbors = tf.sparse.eye(num_vertices, dtype=tf.float64)
    u, v, c, w, b = _random_variables(
        num_channels, num_channels, 1, dtype=np.float64)

    def feature_steered_convolution(data):
      return gc.feature_steered_convolution(
          data=data,
          neighbors=neighbors,
          sizes=None,
          var_u=u,
          var_v=v,
          var_c=c,
          var_w=w,
          var_b=b)

    self.assert_jacobian_is_correct_fn(feature_steered_convolution, [data_init])


class EdgeConvolutionTemplateTests(test_case.TestCase):

  def _zeros(self, vertex_features, _, out_dimensions=None):
    """A callable for `edge_convolution_template`."""
    if out_dimensions is None:
      return tf.zeros_like(vertex_features)
    else:
      return tf.zeros(
          shape=(vertex_features.shape.as_list()[0], out_dimensions),
          dtype=vertex_features.dtype)

  def _pass_through(self, _, neighbor_features):
    """A callable for `edge_convolution_template`."""
    return neighbor_features

  def _circular_2d_data(self, num_vertices, include_normals=False):
    """Create data for a circle graph."""
    # Vertices are points distributed uniformly on a circle, with each point
    # connected to its closest neighbor on either side.
    theta = np.linspace(0.0, np.pi * 2.0, num=num_vertices, endpoint=False)
    data = np.stack((np.cos(theta), np.sin(theta)), axis=-1)
    if include_normals:
      data = np.concatenate((data, data), axis=-1)
    eye = np.eye(num_vertices)
    neighbors = np.maximum(np.roll(eye, 1, axis=1), np.roll(eye, -1,
                                                            axis=1)) * 0.5
    return data, _dense_to_sparse(neighbors)

  def _edge_curvature_2d(self, vertex_features, neighbor_features):
    """A callable for `edge_convolution_template` that computes curvature."""
    x_position, x_normal = tf.split(
        value=vertex_features, num_or_size_splits=2, axis=-1)
    y_position, y_normal = tf.split(
        value=neighbor_features, num_or_size_splits=2, axis=-1)
    yx_diff = x_position - y_position
    curvature_unscaled = tf.abs(
        tf.reduce_sum(
            input_tensor=(y_normal - x_normal) * yx_diff,
            axis=-1,
            keepdims=True))
    edge_length_squared = tf.reduce_sum(
        input_tensor=yx_diff * yx_diff, axis=-1, keepdims=True)
    return tf.compat.v1.where(
        tf.less(edge_length_squared, 1e-7), tf.zeros_like(edge_length_squared),
        curvature_unscaled / edge_length_squared)

  @parameterized.parameters(
      ("'sizes' must have an integer type.", np.float32, np.float32,
       np.float32),
      ("'data' must have a float type.", np.int32, np.float32, np.int32),
      ("'neighbors' and 'data' must have the same type.", np.float32,
       np.float64, np.int32),
  )
  def test_edge_convolution_template_exception_raised_types(
      self, err_msg, data_type, neighbors_type, sizes_type):
    """Check the type errors for invalid input types."""
    data, neighbors, sizes = _random_data(1, 5, 3, True, False, data_type,
                                          neighbors_type, sizes_type)
    with self.assertRaisesRegexp(TypeError, err_msg):
      gc.edge_convolution_template(
          data=data,
          neighbors=neighbors,
          sizes=sizes,
          edge_function=self._zeros,
          reduction="weighted",
          edge_function_kwargs=dict())

  @parameterized.parameters(
      (np.float32, np.float32, np.int32),
      (np.float64, np.float64, np.int32),
      (np.float32, np.float32, np.int64),
      (np.float64, np.float64, np.int64),
      (np.float64, np.float64, np.int8),
      (np.float64, np.float64, np.uint8),
      (np.float64, np.float64, np.int16),
      (np.float64, np.float64, np.uint16),
  )
  def test_edge_convolution_template_exception_not_raised_types(
      self, data_type, neighbors_type, sizes_type):
    """Check there are no exceptions for valid input types."""
    data, neighbors, sizes = _random_data(1, 5, 3, True, False, data_type,
                                          neighbors_type, sizes_type)
    try:
      gc.edge_convolution_template(
          data=data,
          neighbors=neighbors,
          sizes=sizes,
          edge_function=self._zeros,
          reduction="weighted",
          edge_function_kwargs=dict())
    except Exception as e:  # pylint: disable=broad-except
      self.fail("Exception raised: %s" % str(e))

  def test_edge_convolution_template_exception_raised_shapes(self):
    """Check that invalid input shapes trigger the right exceptions."""
    with self.assertRaisesRegexp(ValueError, "must have a rank of 2"):
      data, neighbors = _dummy_data(1, 5, 2)
      data = data[0, :]
      _ = gc.edge_convolution_template(
          data=data,
          neighbors=neighbors,
          sizes=None,
          edge_function=self._zeros,
          reduction="weighted",
          edge_function_kwargs=dict())

    with self.assertRaisesRegexp(ValueError, "must have a rank greater than 1"):
      data = np.ones(shape=(5), dtype=np.float32)
      neighbors = _dense_to_sparse(np.ones(shape=(5), dtype=np.float32))
      _ = gc.edge_convolution_template(
          data=data,
          neighbors=neighbors,
          sizes=None,
          edge_function=self._zeros,
          reduction="weighted",
          edge_function_kwargs=dict())

    with self.assertRaisesRegexp(ValueError, "must have a rank of 1"):
      data, neighbors = _dummy_data(1, 5, 2)
      _ = gc.edge_convolution_template(
          data=data,
          neighbors=neighbors,
          sizes=((1, 1), (1, 1)),
          edge_function=self._zeros,
          reduction="weighted",
          edge_function_kwargs=dict())

  @parameterized.parameters("", "invalid")
  def test_edge_convolution_template_exception_raised_reduction(
      self, reduction):
    """Check that an invalid reduction method triggers the exception."""
    with self.assertRaisesRegexp(ValueError, "reduction method"):
      data, neighbors = _dummy_data(1, 5, 2)
      gc.edge_convolution_template(
          data=data,
          neighbors=neighbors,
          sizes=None,
          edge_function=self._zeros,
          reduction=reduction,
          edge_function_kwargs=dict())

  @parameterized.parameters(
      (1, 1, 1, 1, "weighted"),
      (4, 2, 3, 6, "weighted"),
      (0, 1, 1, 1, "max"),
      (0, 2, 3, 6, "max"),
  )
  def test_edge_convolution_template_output_shape(self, batch_size,
                                                  num_vertices, in_channels,
                                                  out_channels, reduction):
    """Check that the output of convolution has the correct shape."""
    data, neighbors = _dummy_data(batch_size, num_vertices, in_channels)

    y = gc.edge_convolution_template(
        data,
        neighbors,
        None,
        self._zeros,
        reduction=reduction,
        edge_function_kwargs={"out_dimensions": out_channels})
    y_shape = y.shape.as_list()

    with self.subTest(name="out_channels"):
      self.assertEqual(y_shape[-1], out_channels)

    with self.subTest(name="shape"):
      self.assertAllEqual(y_shape[:-1], data.shape[:-1])

  def test_edge_convolution_template_zero_neighbors(self):
    """Check that vertices with no neighbors map to zeros in the output."""
    # We can reuse `self._edge_curvature_2d` as the curvature functional.
    num_vertices = 500
    data, neighbors = self._circular_2d_data(num_vertices, include_normals=True)

    # Interleave the data with rows filled with random data, these rows will
    # have no neighbors in the adjacency matrix so should map to all zeros in
    # the output.
    rows_odd = tf.expand_dims(
        tf.range(start=1, limit=(2 * num_vertices), delta=2), -1)
    rows_even = tf.expand_dims(
        tf.range(start=0, limit=(2 * num_vertices + 1), delta=2), -1)
    data_interleaved = tf.scatter_nd(
        indices=rows_odd,
        updates=data,
        shape=(2 * num_vertices + 1, tf.shape(input=data)[-1]))
    random_data = tf.random.uniform(
        shape=(data.shape[0] + 1, data.shape[-1]), dtype=data.dtype)
    random_interleaved = tf.scatter_nd(
        indices=rows_even,
        updates=random_data,
        shape=(2 * num_vertices + 1, tf.shape(input=data)[-1]))
    data_interleaved = data_interleaved + random_interleaved
    neighbors_interleaved_indices = neighbors.indices * 2 + 1
    neighbors_interleaved = tf.SparseTensor(
        indices=neighbors_interleaved_indices,
        values=neighbors.values,
        dense_shape=(2 * num_vertices + 1, 2 * num_vertices + 1))

    # Convolve the interleaved data.
    data_curvature = gc.edge_convolution_template(
        data=data_interleaved,
        neighbors=neighbors_interleaved,
        sizes=None,
        edge_function=self._edge_curvature_2d,
        reduction="weighted",
        edge_function_kwargs=dict())

    self.assertEqual(data_curvature.shape, (2 * num_vertices + 1, 1))

    # The rows corresponding to the original input data measure the curvature.
    # The curvature at any point on a circle of radius 1 should be 1.
    # The interleaved rows of random data should map to zeros in the output.
    self.assertAllClose(data_curvature[1::2, :],
                        np.ones(shape=(num_vertices, 1)))
    self.assertAllClose(data_curvature[::2, :],
                        np.zeros(shape=(num_vertices + 1, 1)))

  @parameterized.parameters(
      (1, 10, 3, True, "weighted"),
      (3, 6, 1, True, "weighted"),
      (0, 10, 5, False, "weighted"),
      (1, 10, 3, False, "max"),
      (3, 6, 1, False, "max"),
      (0, 10, 5, False, "max"),
  )
  def test_edge_convolution_template_jacobian_random(self, batch_size,
                                                     num_vertices, in_channels,
                                                     padding, reduction):
    """Test the jacobian for random input data."""
    random_data = _random_data(
        batch_size,
        num_vertices,
        in_channels,
        padding,
        only_self_edges=False,
        data_type=np.float64,
        neighbors_type=np.float64)
    data_init = random_data[0]
    neighbors = random_data[1]
    sizes = None if not padding else random_data[2]

    def edge_convolution_template(data):
      return gc.edge_convolution_template(
          data=data,
          neighbors=neighbors,
          sizes=sizes,
          edge_function=self._pass_through,
          reduction=reduction,
          edge_function_kwargs=dict())

    self.assert_jacobian_is_correct_fn(edge_convolution_template, [data_init])

  def test_edge_convolution_template_preset_max(self):
    data = np.array(((1, 2), (3, 4), (5, 6), (7, 8)), np.float32)
    neighbors = np.array(
        ((0, 1, 0, 1), (0, 0, 1, 0), (1, 1, 1, 0), (0, 0, 1, 1)), np.float32)
    neighbors = _dense_to_sparse(neighbors)
    true = np.array(((8, 10), (8, 10), (10, 12), (14, 16)), np.float32)

    with self.subTest("max_sum"):
      max_sum = gc.edge_convolution_template(
          data=data,
          neighbors=neighbors,
          sizes=None,
          edge_function=lambda x, y: x + y,
          reduction="max",
          edge_function_kwargs=dict())

      self.assertAllEqual(max_sum, true)

    with self.subTest("max_sum_scaled"):
      # Max reduction ignores the weights, so scaling the neighbors weights
      # should not change the result.
      max_sum_scaled = gc.edge_convolution_template(
          data=data,
          neighbors=neighbors * 10.0,
          sizes=None,
          edge_function=lambda x, y: x + y,
          reduction="max",
          edge_function_kwargs=dict())

      self.assertAllEqual(max_sum_scaled, true)

  @parameterized.parameters(
      itertools.product((1, 5), (1, 3), (0.0, 1.0), ("weighted", "max")))
  def test_edge_convolution_template_jacobian_preset(self, num_vertices,
                                                     num_channels,
                                                     data_multiplier,
                                                     reduction):
    """Test the jacobian is correct for preset inputs."""
    # Corner cases include one vertex, one channel, and all-zero features.
    data_init = data_multiplier * np.random.uniform(
        size=(num_vertices, num_channels)).astype(np.float64)
    neighbors = tf.sparse.eye(num_vertices, dtype=tf.float64)

    def edge_convolution_template(data):
      return gc.edge_convolution_template(
          data=data,
          neighbors=neighbors,
          sizes=None,
          edge_function=self._pass_through,
          reduction=reduction,
          edge_function_kwargs=dict())

    self.assert_jacobian_is_correct_fn(edge_convolution_template, [data_init])

  def test_edge_convolution_template_laplacian_smoothing(self):
    r"""Test the expected result with laplacian smoothing.

      Laplacian smoothing for meshes is defined as
      $$y_i = \frac{1}{|\mathcal{N(i)}|} \sum_{j \in \mathcal{N(i)}} x_j$$

      This can be computed using `edge_convolution_template` with `f(x, y)->y`.
    """

    # We can reuse `self._pass_through(x, y)->y` as the smoothing functional.
    with self.subTest(name="only_self_edges_random"):
      num_vertices = 500
      data = np.random.uniform(size=(num_vertices, 5))
      neighbors = tf.sparse.eye(num_vertices, dtype=tf.as_dtype(data.dtype))

      data_smoothed = gc.edge_convolution_template(
          data=data,
          neighbors=neighbors,
          sizes=None,
          edge_function=self._pass_through,
          reduction="weighted",
          edge_function_kwargs=dict())

      self.assertAllEqual(data, data_smoothed)

    with self.subTest(name="circular_2d"):
      num_vertices = 500
      data, neighbors = self._circular_2d_data(num_vertices)

      data_smoothed = gc.edge_convolution_template(
          data=data,
          neighbors=neighbors,
          sizes=None,
          edge_function=self._pass_through,
          reduction="weighted",
          edge_function_kwargs=dict())
      # The smoothed points should have the same direction as the originals.
      data_smoothed_normalized = tf.nn.l2_normalize(data_smoothed, axis=-1)

      self.assertAllClose(data, data_smoothed_normalized)

  def test_edge_convolution_template_curvature(self):
    r"""Test the expected result with curvature.

      (Approximate) curvature for meshes is defined as
      $$\kappa_{v_i} = \frac{1}{|\mathcal{N}(v_i)|}
        \sum_{v_j \in \mathcal{N}(v_i)}
        \frac{(\vec{v_i} - \vec{v_j})^T (\vec{n_{v_i}} -
        \vec{n_{v_j}})} {\left|\vec{v_i}-\vec{v_j}\right|^2}
      $$

      This can be computed using `edge_convolution_template` with
        $$f(x, y) = (n_x - n_y)^T (x - y) / ||x - y||^2.$$
      where $$n_x$$ and $$n_y$$ are the normals at points $$x$$ and $$y$$
      respectively.
    """
    # We can reuse `self._edge_curvature_2d` as the curvature functional.
    num_vertices = 500
    data, neighbors = self._circular_2d_data(num_vertices, include_normals=True)

    data_curvature = gc.edge_convolution_template(
        data=data,
        neighbors=neighbors,
        sizes=None,
        edge_function=self._edge_curvature_2d,
        reduction="weighted",
        edge_function_kwargs=dict())

    # The curvature at each point on a circle of radius 1 should be 1.
    self.assertAllClose(data_curvature, np.ones(shape=(num_vertices, 1)))


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for keras layers."""

import tensorflow as tf

layers = tf.keras.layers
initializer = tf.keras.initializers.glorot_normal()
rate = 0.7


def norm_layer(tensor, normalization):
  if normalization.lower() == 'batchnorm':
    tensor = layers.BatchNormalization()(tensor)
  return tensor


def upconv(tensor, nfilters, size, strides,
           alpha_lrelu=0.2, normalization='None'):
  """Upconvolution as upsampling and convolution."""
  tensor = layers.UpSampling2D()(tensor)
  tensor = layers.Conv2D(nfilters, size,
                         strides=strides,
                         padding='same',
                         kernel_initializer=initializer,
                         use_bias=False)(tensor)

  tensor = norm_layer(tensor, normalization)
  tensor = layers.LeakyReLU(alpha=alpha_lrelu)(tensor)
  if normalization.lower() == 'dropout':
    tensor = layers.Dropout(rate)(tensor)
  return tensor


def conv_block_3d(tensor, nfilters, size, strides,
                  alpha_lrelu=0.2, normalization='None', relu=True):
  """3D convolution block with normalization and leaky relu."""
  tensor = layers.Conv3D(nfilters, size,
                         strides=strides,
                         padding='same',
                         kernel_initializer=initializer,
                         use_bias=False)(tensor)

  tensor = norm_layer(tensor, normalization)

  if relu:
    tensor = layers.LeakyReLU(alpha=alpha_lrelu)(tensor)
  if normalization.lower() == 'dropout':
    tensor = layers.Dropout(rate)(tensor)
  return tensor


def conv_t_block_3d(tensor, nfilters, size, strides,
                    alpha_lrelu=0.2, normalization='None', relu=True):
  """2D transpose convolution block with normalization and leaky relu."""
  tensor = layers.Conv3DTranspose(nfilters, size,
                                  strides=strides,
                                  padding='same',
                                  kernel_initializer=initializer,
                                  use_bias=False)(tensor)

  tensor = norm_layer(tensor, normalization)
  if relu:
    tensor = layers.LeakyReLU(alpha=alpha_lrelu)(tensor)
  if normalization.lower() == 'dropout':
    tensor = layers.Dropout(rate)(tensor)
  return tensor


def conv_block_2d(tensor, nfilters, size, strides,
                  alpha_lrelu=0.2, normalization='None'):
  """2D convolution block with normalization and leaky relu."""
  tensor = layers.Conv2D(nfilters, size,
                         strides=strides,
                         padding='same',
                         kernel_initializer=initializer,
                         use_bias=False)(tensor)

  tensor = norm_layer(tensor, normalization)

  tensor = layers.LeakyReLU(alpha=alpha_lrelu)(tensor)
  if normalization.lower() == 'dropout':
    tensor = layers.Dropout(rate)(tensor)
  return tensor


def conv_t_block_2d(tensor, nfilters, size, strides,
                    alpha_lrelu=0.2, normalization='None'):
  """2D transpose convolution block with normalization and leaky relu."""
  tensor = layers.Conv2DTranspose(nfilters, size,
                                  strides=strides,
                                  padding='same',
                                  kernel_initializer=initializer,
                                  use_bias=False)(tensor)

  tensor = norm_layer(tensor, normalization)

  tensor = layers.LeakyReLU(alpha=alpha_lrelu)(tensor)
  if normalization.lower() == 'dropout':
    tensor = layers.Dropout(rate)(tensor)
  return tensor


def residual_block_2d(x, nfilters, strides=(1, 1), normalization='None'):
  """2D residual block."""
  shortcut = x

  x = layers.Conv2D(nfilters,
                    kernel_size=(3, 3),
                    strides=strides,
                    padding='same',
                    kernel_initializer=initializer)(x)
  x = norm_layer(x, normalization)
  x = layers.LeakyReLU()(x)

  x = layers.Conv2D(nfilters,
                    kernel_size=(3, 3),
                    strides=(1, 1),
                    padding='same',
                    kernel_initializer=initializer)(x)
  x = norm_layer(x, normalization)

  if strides != (1, 1):
    shortcut = layers.Conv2D(nfilters,
                             kernel_size=(1, 1),
                             strides=strides,
                             padding='same')(shortcut)
    x = norm_layer(x, normalization)

  x = layers.add([shortcut, x])
  x = layers.LeakyReLU()(x)

  return x


def residual_block_3d(x, nfilters, strides=(1, 1, 1), normalization='None'):
  """3D residual block."""
  shortcut = x

  x = layers.Conv3D(nfilters,
                    kernel_size=(3, 3, 3),
                    strides=strides,
                    padding='same',
                    kernel_initializer=initializer)(x)
  x = norm_layer(x, normalization)
  x = layers.LeakyReLU()(x)

  x = layers.Conv3D(nfilters,
                    kernel_size=(3, 3, 3),
                    strides=(1, 1, 1),
                    padding='same',
                    kernel_initializer=initializer)(x)
  x = norm_layer(x, normalization)

  if strides != (1, 1, 1):
    shortcut = layers.Conv3D(nfilters,
                             kernel_size=(1, 1, 1),
                             strides=strides,
                             padding='same')(shortcut)
    x = norm_layer(x, normalization)

  x = layers.add([shortcut, x])
  x = layers.LeakyReLU()(x)

  return x



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for orthographic camera functionalities."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import numpy as np

from tensorflow_graphics.rendering.camera import orthographic
from tensorflow_graphics.util import test_case


class OrthographicTest(test_case.TestCase):

  @parameterized.parameters(
      ((3,),),
      ((None, 3),),
  )
  def test_project_exception_not_exception_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(orthographic.project, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (1,)),
      ("must have exactly 3 dimensions in axis -1", (2,)),
      ("must have exactly 3 dimensions in axis -1", (4,)),
  )
  def test_project_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(orthographic.project, error_msg, shape)

  def test_project_jacobian_random(self):
    """Test the Jacobian of the project function."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    point_3d_init = np.random.uniform(size=tensor_shape + [3])

    self.assert_jacobian_is_correct_fn(orthographic.project, [point_3d_init])

  def test_project_random(self):
    """Test the project function using random 3d points."""
    point_3d = np.random.uniform(size=(100, 3))

    pred = orthographic.project(point_3d)

    self.assertAllEqual(pred, point_3d[:, 0:2])

  @parameterized.parameters(
      ((2,),),
      ((None, 2),),
  )
  def test_ray_exception_not_exception_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(orthographic.ray, shapes)

  @parameterized.parameters(
      ("must have exactly 2 dimensions in axis -1", (1,)),
      ("must have exactly 2 dimensions in axis -1", (3,)),
  )
  def test_ray_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(orthographic.ray, error_msg, shape)

  def test_ray_jacobian_random(self):
    """Test the Jacobian of the ray function."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    point_2d_init = np.random.uniform(size=tensor_shape + [2])

    self.assert_jacobian_is_correct_fn(orthographic.ray, [point_2d_init])

  def test_ray_random(self):
    """Test the ray function using random 2d points."""
    point_2d = np.random.uniform(size=(100, 2))

    pred = orthographic.ray(point_2d)
    gt = np.tile((0.0, 0.0, 1.0), (100, 1))

    self.assertAllEqual(pred, gt)

  @parameterized.parameters(
      ((2,), (1,)),
      ((None, 2), (None, 1)),
      ((2, 2), (2, 1)),
  )
  def test_unproject_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(orthographic.unproject, shapes)

  @parameterized.parameters(
      ("must have exactly 2 dimensions in axis -1", (1,), (1,)),
      ("must have exactly 2 dimensions in axis -1", (3,), (1,)),
      ("must have exactly 1 dimensions in axis -1", (2,), (2,)),
      ("Not all batch dimensions are identical.", (3, 2), (2, 1)),
  )
  def test_unproject_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(orthographic.unproject, error_msg, shape)

  def test_unproject_jacobian_random(self):
    """Test the Jacobian of the unproject function."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    point_2d_init = np.random.uniform(size=tensor_shape + [2])
    depth_init = np.random.uniform(size=tensor_shape + [1])

    self.assert_jacobian_is_correct_fn(orthographic.unproject,
                                       [point_2d_init, depth_init])

  def test_unproject_random(self):
    """Test the unproject function using random 2d points."""
    point_2d = np.random.uniform(size=(100, 2))
    depth = np.random.uniform(size=(100, 1))

    pred = orthographic.unproject(point_2d, depth)

    self.assertAllEqual(pred[:, 0:2], point_2d)
    self.assertAllEqual(pred[:, 2], np.squeeze(depth))


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for matting."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.image import matting
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import shape
from tensorflow_graphics.util import test_case


def _laplacian_matrix(image, size=3, eps=1e-5, name=None):
  """Generates the closed form matting Laplacian matrices.

  Generates the closed form matting Laplacian as proposed by Levin et
  al. in "A Closed Form Solution to Natural Image Matting".

  Args:
    image: A tensor of shape `[B, H, W, C]`.
    size: An `int` representing the size of the patches used to enforce
      smoothness.
    eps: A small number of type `float` to regularize the problem.
    name: A name for this op. Defaults to "matting_laplacian_matrix".

  Returns:
    A tensor of shape `[B, H, W, size^2, size^2]` containing the
    matting Laplacian matrices.

  Raises:
    ValueError: If `image` is not of rank 4.
  """
  with tf.compat.v1.name_scope(name, "matting_laplacian_matrix", [image]):
    image = tf.convert_to_tensor(value=image)

    shape.check_static(image, has_rank=4)
    if size % 2 == 0:
      raise ValueError("The patch size is expected to be an odd value.")

    pixels = size**2
    channels = tf.shape(input=image)[-1]
    dtype = image.dtype
    patches = tf.image.extract_patches(
        image,
        sizes=(1, size, size, 1),
        strides=(1, 1, 1, 1),
        rates=(1, 1, 1, 1),
        padding="VALID")
    batches = tf.shape(input=patches)[:-1]
    new_shape = tf.concat((batches, (pixels, channels)), axis=-1)
    patches = tf.reshape(patches, shape=new_shape)
    mean = tf.reduce_mean(input_tensor=patches, axis=-2, keepdims=True)
    demean = patches - mean
    covariance = tf.matmul(demean, demean, transpose_a=True) / pixels
    regularizer = (eps / pixels) * tf.eye(channels, dtype=dtype)
    covariance_inv = tf.linalg.inv(covariance + regularizer)
    covariance_inv = asserts.assert_no_infs_or_nans(covariance_inv)
    mat = tf.matmul(tf.matmul(demean, covariance_inv), demean, transpose_b=True)
    return tf.eye(pixels, dtype=dtype) - (1.0 + mat) / pixels


class MattingTest(test_case.TestCase):

  @parameterized.parameters((3, 1), (3, 3), (5, 3), (5, 1))
  def test_build_matrices_jacobian_random(self, size, channels):
    """Tests the Jacobian of the build_matrices function."""
    tensor_shape = np.random.randint(size, 6, size=3)
    image_init = np.random.uniform(
        0.0, 1.0, size=tensor_shape.tolist() + [channels])

    with self.subTest(name="laplacian"):
      self.assert_jacobian_is_correct_fn(
          lambda image: matting.build_matrices(image, size=size)[0],
          [image_init])
    with self.subTest(name="pseudo_inverse"):
      self.assert_jacobian_is_correct_fn(
          lambda image: matting.build_matrices(image, size=size)[1],
          [image_init])

  @parameterized.parameters((3, 1), (3, 3), (5, 3), (5, 1))
  def test_build_matrices_laplacian_zero_rows_and_columns(self, size, channels):
    """Tests that the laplacian matrix rows and columns sum to zero."""
    tensor_shape = np.random.randint(size, 6, size=3)
    image_init = np.random.uniform(
        0.0, 1.0, size=tensor_shape.tolist() + [channels])
    image = tf.convert_to_tensor(value=image_init)

    laplacian, _ = matting.build_matrices(image, size=size)
    rows = tf.reduce_sum(input_tensor=laplacian, axis=-2)
    columns = tf.reduce_sum(input_tensor=laplacian, axis=-1)

    with self.subTest(name="rows"):
      self.assertAllClose(rows, tf.zeros_like(rows))
    with self.subTest(name="columns"):
      self.assertAllClose(columns, tf.zeros_like(columns))

  @parameterized.parameters((3, 1), (3, 3), (5, 3), (5, 1))
  def test_build_matrices_laplacian_versions(self, size, channels):
    """Compares two ways of computing the laplacian matrix."""
    tensor_shape = np.random.randint(size, 6, size=3)
    image_init = np.random.uniform(
        0.0, 1.0, size=tensor_shape.tolist() + [channels])
    image = tf.convert_to_tensor(value=image_init)

    laplacian_v1, _ = matting.build_matrices(image, size=size)
    laplacian_v2 = _laplacian_matrix(image, size=size)

    self.assertAllClose(laplacian_v1, laplacian_v2)

  @parameterized.parameters(
      (3, (None, None, None, 1)),
      (3, (None, None, None, 3)),
      (5, (None, None, None, 1)),
      (5, (None, None, None, 3)),
      (3, (1, 3, 3, 1)),
      (3, (1, 3, 3, 3)),
      (5, (1, 5, 5, 1)),
      (5, (1, 5, 5, 3)),
  )
  def test_build_matrices_not_raised(self, size, *shapes):
    """Tests that the shape exceptions are not raised."""
    build_matrices = lambda image: matting.build_matrices(image, size=size)
    self.assert_exception_is_not_raised(build_matrices, shapes)

  @parameterized.parameters(
      ("tensor must have a rank of 4, but it has rank", 3, (1,)),
      ("tensor must have a rank of 4, but it has rank", 3, (1, 1, 1, 1, 1)),
      ("The patch size is expected to be an odd value.", 2, (1, 1, 1, 1)),
  )
  def test_build_matrices_raised(self, error_msg, size, *shapes):
    """Tests that the shape exceptions are properly raised."""
    build_matrices = lambda image: matting.build_matrices(image, size=size)
    self.assert_exception_is_raised(build_matrices, error_msg, shapes)

  @parameterized.parameters((3,), (5,))
  def test_linear_coefficients_jacobian_random(self, size):
    """Tests the Jacobian of the linear_coefficients function."""
    tensor_shape = np.random.randint(size, 6, size=3)
    matte_init = np.random.uniform(0.0, 1.0, size=tensor_shape.tolist() + [1])
    tensor_shape[1:3] -= (size - 1)
    num_coeffs = np.random.randint(2, 4)
    pseudo_inverse_init = np.random.uniform(
        0.0, 1.0, size=tensor_shape.tolist() + [num_coeffs, size**2])

    def a_fn(matte, pseudo_inverse):
      a, _ = matting.linear_coefficients(matte, pseudo_inverse)
      return a

    def b_fn(matte, pseudo_inverse):
      _, b = matting.linear_coefficients(matte, pseudo_inverse)
      return b

    with self.subTest(name="a"):
      self.assert_jacobian_is_correct_fn(a_fn,
                                         [matte_init, pseudo_inverse_init])
    with self.subTest(name="b"):
      self.assert_jacobian_is_correct_fn(b_fn,
                                         [matte_init, pseudo_inverse_init])

  @parameterized.parameters(
      ((None, None, None, 1), (None, None, None, 4, 9)),
      ((None, None, None, 1), (None, None, None, 2, 25)),
      ((1, 6, 6, 1), (1, 4, 4, 2, 9)),
      ((1, 10, 10, 1), (1, 6, 6, 2, 25)),
  )
  def test_linear_coefficients_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(matting.linear_coefficients, shapes)

  @parameterized.parameters(
      ("must have exactly 1 dimensions in axis -1", (1, 6, 6, 2),
       (1, 4, 4, 2, 9)),
      ("Not all batch dimensions are identical.", (1, 6, 6, 1),
       (2, 4, 4, 2, 9)),
  )
  def test_linear_coefficients_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(matting.linear_coefficients, error_msg,
                                    shapes)

  @parameterized.parameters((3,), (5,))
  def test_linear_coefficients_reconstruction_same_images(self, size):
    """Tests that the matte can be reconstructed by using the coefficients ."""
    tensor_shape = np.random.randint(size, 6, size=3).tolist()
    image = np.random.uniform(0.0, 1.0, size=tensor_shape + [1])

    _, pseudo_inverse = matting.build_matrices(image, size=size)
    a, b = matting.linear_coefficients(image, pseudo_inverse)
    reconstructed = matting.reconstruct(image, a, b)

    self.assertAllClose(image, reconstructed, atol=1e-4)

  @parameterized.parameters((3,), (5,))
  def test_linear_coefficients_reconstruction_opposite_images(self, size):
    """Tests that the matte can be reconstructed by using the coefficients ."""
    tensor_shape = np.random.randint(size, 6, size=3).tolist()
    image = np.random.uniform(0.0, 1.0, size=tensor_shape + [1])

    _, pseudo_inverse = matting.build_matrices(image, size=size)
    a, b = matting.linear_coefficients(1.0 - image, pseudo_inverse)
    reconstructed = matting.reconstruct(image, a, b)

    self.assertAllClose(1.0 - image, reconstructed, atol=1e-4)

  @parameterized.parameters((3,), (5,))
  def test_loss_jacobian_random(self, size):
    """Tests the Jacobian of the matting loss function."""
    tensor_shape = np.random.randint(size, 6, size=3)
    matte_init = np.random.uniform(0.0, 1.0, size=tensor_shape.tolist() + [1])
    tensor_shape[1:3] -= (size - 1)
    laplacian_init = np.random.uniform(
        0.0, 1.0, size=tensor_shape.tolist() + [size**2, size**2])

    with self.subTest(name="matte"):
      self.assert_jacobian_is_correct_fn(matting.loss,
                                         [matte_init, laplacian_init])

  @parameterized.parameters(
      ((None, None, None, 1), (None, None, None, 9, 9)),
      ((None, None, None, 1), (None, None, None, 25, 25)),
      ((1, 6, 6, 1), (1, 4, 4, 9, 9)),
      ((1, 10, 10, 1), (1, 6, 6, 25, 25)),
  )
  def test_loss_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(matting.loss, shapes)

  @parameterized.parameters(
      ("must have exactly 1 dimensions in axis -1", (1, 6, 6, 2),
       (1, 4, 4, 9, 9)),
      ("must have exactly 9 dimensions in axis -2", (1, 6, 6, 1),
       (1, 4, 4, 1, 9)),
      ("Not all batch dimensions are identical.", (1, 6, 6, 1),
       (2, 4, 4, 9, 9)),
  )
  def test_loss_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(matting.loss, error_msg, shapes)

  @parameterized.parameters((3,), (5,))
  def test_loss_opposite_images(self, size):
    """Tests that passing opposite images results in a loss close to 0.0."""
    tensor_shape = np.random.randint(size, 6, size=3).tolist()
    image = np.random.uniform(0.0, 1.0, size=tensor_shape + [1])

    laplacian, _ = matting.build_matrices(image, size=size)
    loss = matting.loss(1.0 - image, laplacian)

    self.assertAllClose(loss, 0.0, atol=1e-4)

  @parameterized.parameters((3,), (5,))
  def test_loss_same_images(self, size):
    """Tests that passing same images results in a loss close to 0.0."""
    tensor_shape = np.random.randint(size, 6, size=3).tolist()
    image = np.random.uniform(0.0, 1.0, size=tensor_shape + [1])

    laplacian, _ = matting.build_matrices(image, size=size)
    loss = matting.loss(image, laplacian)

    self.assertAllClose(loss, 0.0, atol=1e-4)

  @parameterized.parameters((3,), (5,))
  def test_loss_positive(self, size):
    """Tests that the loss is always greater or equal to 0.0."""
    tensor_shape = np.random.randint(size, 6, size=3).tolist()
    image = tf.random.uniform(minval=0.0, maxval=1.0, shape=tensor_shape + [3])
    matte = tf.random.uniform(minval=0.0, maxval=1.0, shape=tensor_shape + [1])

    laplacian, _ = matting.build_matrices(image, size=size)
    loss = matting.loss(matte, laplacian)

    self.assertAllGreaterEqual(loss, 0.0)

  @parameterized.parameters((1,), (3,))
  def test_reconstruct_jacobian_random(self, channels):
    """Tests the Jacobian of the reconstruct function."""
    tensor_shape = np.random.randint(1, 5, size=3).tolist()
    image_init = np.random.uniform(0.0, 1.0, size=tensor_shape + [channels])
    mul_init = np.random.uniform(0.0, 1.0, size=tensor_shape + [channels])
    add_init = np.random.uniform(0.0, 1.0, size=tensor_shape + [1])

    self.assert_jacobian_is_correct_fn(matting.reconstruct,
                                       [image_init, mul_init, add_init])

  @parameterized.parameters(
      ((None, None, None, 3), (None, None, None, 3), (None, None, None, 1)),
      ((1, 6, 6, 3), (1, 6, 6, 3), (1, 6, 6, 1)),
  )
  def test_reconstruct_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(matting.reconstruct, shapes)

  @parameterized.parameters(
      ("tensor must have a rank of 4, but it has rank", (1, 6, 6), (1, 6, 6, 2),
       (1, 6, 6, 1)),
      ("tensor must have a rank of 4, but it has rank", (1, 6, 6, 2), (1, 6, 6),
       (1, 6, 6, 1)),
      ("tensor must have a rank of 4, but it has rank", (1, 6, 6, 2),
       (1, 6, 6, 2), (1, 6, 6)),
      ("must have exactly 1 dimensions in axis -1", (1, 6, 6, 2), (1, 6, 6, 2),
       (1, 6, 6, 2)),
      ("Not all batch dimensions are identical.", (1, 6, 6, 1), (1, 6, 6, 4),
       (1, 6, 6, 1)),
      ("Not all batch dimensions are identical.", (1, 6, 6, 1), (1, 4, 6, 1),
       (1, 6, 6, 1)),
      ("Not all batch dimensions are identical.", (1, 6, 6, 1), (1, 6, 6, 1),
       (1, 4, 6, 1)),
      ("Not all batch dimensions are identical.", (1, 6, 6, 1), (1, 6, 4, 1),
       (1, 6, 6, 1)),
      ("Not all batch dimensions are identical.", (1, 6, 6, 1), (1, 6, 6, 1),
       (1, 6, 4, 1)),
      ("Not all batch dimensions are identical.", (1, 6, 6, 1), (4, 6, 6, 1),
       (1, 6, 6, 1)),
      ("Not all batch dimensions are identical.", (1, 6, 6, 1), (1, 6, 6, 1),
       (4, 6, 6, 1)),
  )
  def test_reconstruct_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(matting.reconstruct, error_msg, shapes)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for google3.third_party.py.tensorflow_graphics.interpolation.weighted."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.math.interpolation import weighted
from tensorflow_graphics.util import test_case


class WeightedTest(test_case.TestCase):

  def _get_tensors_from_shapes(self, num_points, dim_points, num_outputs,
                               num_pts_to_interpolate):
    points = np.random.uniform(size=(num_points, dim_points))
    weights = np.random.uniform(size=(num_outputs, num_pts_to_interpolate))
    indices = np.asarray([
        np.random.permutation(num_points)[:num_pts_to_interpolate].tolist()
        for _ in range(num_outputs)
    ])
    indices = np.expand_dims(indices, axis=-1)
    return points, weights, indices

  @parameterized.parameters(
      (3, 4, 2, 3),
      (5, 4, 5, 3),
      (5, 6, 5, 5),
      (2, 6, 5, 1),
  )
  def test_interpolate_exception_not_raised(self, dim_points, num_points,
                                            num_outputs,
                                            num_pts_to_interpolate):
    """Tests whether exceptions are not raised for compatible shapes."""
    points, weights, indices = self._get_tensors_from_shapes(
        num_points, dim_points, num_outputs, num_pts_to_interpolate)

    self.assert_exception_is_not_raised(
        weighted.interpolate,
        shapes=[],
        points=points,
        weights=weights,
        indices=indices,
        normalize=True)

  @parameterized.parameters(
      ("must have a rank greater than 1", ((3,), (None, 2), (None, 2, 0))),
      ("must have a rank greater than 1", ((None, 3), (None, 2), (1,))),
      ("must have exactly 1 dimensions in axis -1", ((None, 3), (None, 2),
                                                     (None, 2, 2))),
      ("must have the same number of dimensions", ((None, 3), (None, 2),
                                                   (None, 3, 1))),
      ("Not all batch dimensions are broadcast-compatible.",
       ((None, 3), (None, 5, 2), (None, 4, 2, 1))),
  )
  def test_interpolate_exception_raised(self, error_msg, shapes):
    """Tests whether exceptions are raised for incompatible shapes."""
    self.assert_exception_is_raised(
        weighted.interpolate, error_msg, shapes=shapes, normalize=False)

  @parameterized.parameters(
      (((-1.0, 1.0), (1.0, 1.0), (3.0, 1.0), (-1.0, -1.0), (1.0, -1.0),
        (3.0, -1.0)), ((0.25, 0.25, 0.25, 0.25), (0.5, 0.5, 0.0, 0.0)),
       (((0,), (1,), (3,), (4,)), ((1,), (2,), (4,),
                                   (5,))), False, ((0.0, 0.0), (2.0, 1.0))),)
  def test_interpolate_preset(self, points, weights, indices, _, out):
    """Tests whether interpolation results are correct."""
    weights = tf.convert_to_tensor(value=weights)

    result_unnormalized = weighted.interpolate(
        points=points, weights=weights, indices=indices, normalize=False)
    result_normalized = weighted.interpolate(
        points=points, weights=2.0 * weights, indices=indices, normalize=True)
    estimated_unnormalized = self.evaluate(result_unnormalized)
    estimated_normalized = self.evaluate(result_normalized)

    self.assertAllClose(estimated_unnormalized, out)
    self.assertAllClose(estimated_normalized, out)

  @parameterized.parameters(
      (3, 4, 2, 3),
      (5, 4, 5, 3),
      (5, 6, 5, 5),
      (2, 6, 5, 1),
  )
  def test_interpolate_negative_weights_raised(self, dim_points, num_points,
                                               num_outputs,
                                               num_pts_to_interpolate):
    """Tests whether exception is raised when weights are negative."""
    points, weights, indices = self._get_tensors_from_shapes(
        num_points, dim_points, num_outputs, num_pts_to_interpolate)
    weights *= -1.0

    with self.assertRaises(tf.errors.InvalidArgumentError):
      result = weighted.interpolate(
          points=points, weights=weights, indices=indices, normalize=True)
      self.evaluate(result)

  @parameterized.parameters(
      (((-1.0, 1.0), (1.0, 1.0), (3.0, 1.0), (-1.0, -1.0), (1.0, -1.0),
        (3.0, -1.0)), ((1.0, -1.0, 1.0, -1.0), (0.0, 0.0, 0.0, 0.0)),
       (((0,), (1,), (3,), (4,)), ((1,), (2,), (4,), (5,))), ((0.0, 0.0),
                                                              (0.0, 0.0))))
  def test_interp_unnormalizable_raised_(self, points, weights, indices, _):
    """Tests whether exception is raised when weights are unnormalizable."""
    with self.assertRaises(tf.errors.InvalidArgumentError):
      result = weighted.interpolate(
          points=points,
          weights=weights,
          indices=indices,
          normalize=True,
          allow_negative_weights=True)
      self.evaluate(result)

  @parameterized.parameters(
      (3, 4, 2, 3),
      (5, 4, 5, 3),
      (5, 6, 5, 5),
      (2, 6, 5, 1),
  )
  def test_interpolate_jacobian_random(self, dim_points, num_points,
                                       num_outputs, num_pts_to_interpolate):
    """Tests whether jacobian is correct."""
    points_np, weights_np, indices_np = self._get_tensors_from_shapes(
        num_points, dim_points, num_outputs, num_pts_to_interpolate)

    def interpolate_fn(points, weights):
      return weighted.interpolate(
          points=points, weights=weights, indices=indices_np, normalize=True)

    self.assert_jacobian_is_correct_fn(interpolate_fn, [points_np, weights_np])

  @parameterized.parameters(
      ((3, 2), (2, 2)),
      ((None, 3, 2), (None, 1, 2)),
      ((10, 5, 3, 2), (10, 5, 2, 2)),
  )
  def test_get_barycentric_coordinates_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(weighted.get_barycentric_coordinates,
                                        shapes)

  @parameterized.parameters(
      ("triangle_vertices must have exactly 2 dimensions in axis -1", (3, 1),
       (1, 2)),
      ("triangle_vertices must have exactly 3 dimensions in axis -2", (2, 2),
       (1, 2)),
      ("pixels must have exactly 2 dimensions in axis -1", (3, 2), (1, 3)),
      ("Not all batch dimensions are broadcast-compatible", (5, 3, 2),
       (2, 10, 2)),
  )
  def test_get_barycentric_coordinates_exception_raised(self, error_msg,
                                                        *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(weighted.get_barycentric_coordinates,
                                    error_msg, shape)

  def test_get_barycentric_coordinates_jacobian_random(self):
    """Tests the Jacobian of get_barycentric_coordinates."""
    tensor_size = np.random.randint(2)
    tensor_shape = np.random.randint(1, 2, size=(tensor_size)).tolist()
    triangle_vertices_init = 0.4 * np.random.random(
        tensor_shape + [3, 2]).astype(np.float64) - 0.2
    triangle_vertices_init += np.array(
        ((0.25, 0.25), (0.5, 0.75), (0.75, 0.25)))
    pixels_init = np.random.random(tensor_shape + [3, 2]).astype(np.float64)

    barycentric_fn = weighted.get_barycentric_coordinates

    self.assert_jacobian_is_correct_fn(
        lambda vertices, pixels: barycentric_fn(vertices, pixels)[0],
        [triangle_vertices_init, pixels_init])

  def test_get_barycentric_coordinates_normalized(self):
    """Tests whether the barycentric coordinates are normalized."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    num_pixels = np.random.randint(1, 10)
    pixels_shape = tensor_shape + [num_pixels]
    triangle_vertices = np.random.random(tensor_shape + [3, 2])
    pixels = np.random.random(pixels_shape + [2])

    barycentric_coordinates, _ = weighted.get_barycentric_coordinates(
        triangle_vertices, pixels)
    barycentric_coordinates_sum = tf.reduce_sum(
        input_tensor=barycentric_coordinates, axis=-1)

    self.assertAllClose(barycentric_coordinates_sum, np.full(pixels_shape, 1.0))


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for math_helpers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.math import math_helpers
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import test_case


class MathTest(test_case.TestCase):

  @parameterized.parameters(
      (((0.0, 0.0, 0.0),), ((0.0, np.pi / 2.0, 0.0),)),
      (((2.0, 0.0, 0.0),), ((2.0, np.pi / 2.0, 0.0),)),
      (((0.0, 1.0, 0.0),), ((1.0, np.pi / 2.0, np.pi / 2.0),)),
      (((0.0, 0.0, 1.0),), ((1.0, 0.0, 0.0),)),
      (((-1.0, 0.0, 0.0),), ((1.0, np.pi / 2.0, np.pi),)),
      (((0.0, -1.0, 0.0),), ((1.0, np.pi / 2.0, -np.pi / 2.0),)),
      (((0.0, 0.0, -1.0),), ((1.0, np.pi, 0.0),)),
  )
  def test_cartesian_to_spherical_coordinates_preset(self, test_inputs,
                                                     test_outputs):
    """Tests that cartesian_to_spherical_coordinates behaves as expected."""
    self.assert_output_is_correct(
        math_helpers.cartesian_to_spherical_coordinates, test_inputs,
        test_outputs)

  @parameterized.parameters(
      ((3,),),
      ((None, 3),),
  )
  def test_cartesian_to_spherical_coordinates_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        math_helpers.cartesian_to_spherical_coordinates, shape)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (1,)),)
  def test_cartesian_to_spherical_coordinates_raised(self, error_msg, *shape):
    """Tests that the shape exception is raised."""
    self.assert_exception_is_raised(
        math_helpers.cartesian_to_spherical_coordinates, error_msg, shape)

  def test_cartesian_to_spherical_coordinates_jacobian_random(self):
    """Test the Jacobian of the spherical_to_cartesian_coordinates function."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    point_init = np.random.uniform(-10.0, 10.0, size=tensor_shape + [3])

    self.assert_jacobian_is_correct_fn(
        math_helpers.cartesian_to_spherical_coordinates, [point_init])

  @parameterized.parameters(
      (((1.0, 1.0, 1.0),),),
      (((1.0, 0.0, 0.0),),),
      (((0.0, 1.0, 0.0),),),
  )
  def test_cartesian_to_spherical_coordinates_jacobian_preset(self, cartesian):
    """Test the Jacobian of the spherical_to_cartesian_coordinates function."""
    point_init = np.asarray(cartesian)

    self.assert_jacobian_is_correct_fn(
        math_helpers.cartesian_to_spherical_coordinates, [point_init])

  @parameterized.parameters(
      (((1.0, 1.0, 0.0),), ((np.sqrt(2.0), np.pi / 2.0, np.pi / 4.0),)),
      (((1.0, 0.0, 0.0),), ((1.0, np.pi / 2.0, 0.0),)),
      (((0.0, 1.0, 0.0),), ((1.0, np.pi / 2.0, np.pi / 2.0),)),
      (((0.0, 0.0, 1.0),), ((1.0, 0.0, 0.0),)),
      (((0.0, 0.0, 0.0),), ((0.0, np.pi / 2.0, 0.0),)),
  )
  def test_cartesian_to_spherical_coordinates_values_preset(
      self, test_inputs, test_outputs):
    """Test the Jacobian of the spherical_to_cartesian_coordinates function."""
    self.assert_output_is_correct(
        math_helpers.cartesian_to_spherical_coordinates, test_inputs,
        test_outputs)

  @parameterized.parameters(
      (((0, 1, 5, 6, 15.0),), ((1, 1, 15, 48, 2027025.0),)),)
  def test_double_factorial_preset(self, test_inputs, test_outputs):
    """Tests that double_factorial generates expected results."""
    self.assert_output_is_correct(math_helpers.double_factorial, test_inputs,
                                  test_outputs)

  @parameterized.parameters(
      (((0, 1, 2, 3, 4.0),), ((1, 1, 2, 6, 24.0),)),)
  def test_factorial_preset(self, test_inputs, test_outputs):
    """Tests that double_factorial generates expected results."""
    self.assert_output_is_correct(math_helpers.factorial, test_inputs,
                                  test_outputs)

  @parameterized.parameters(
      (((2.0, np.pi / 2.0, 0.0),), ((2.0, 0.0, 0.0),)),
      (((2.0, -3.0 * np.pi / 2.0, 0.0),), ((2.0, 0.0, 0.0),)),
      (((1.0, np.pi / 2.0, np.pi / 2.0),), ((0.0, 1.0, 0.0),)),
      (((1.0, 0.0, 0.0),), ((0.0, 0.0, 1.0),)),
  )
  def test_spherical_to_cartesian_coordinates_preset(self, test_inputs,
                                                     test_outputs):
    """Tests that spherical_to_cartesian_coordinates behaves as expected."""
    self.assert_output_is_correct(
        math_helpers.spherical_to_cartesian_coordinates, test_inputs,
        test_outputs)

  @parameterized.parameters(
      ((3,),),
      ((None, 3),),
  )
  def test_spherical_to_cartesian_coordinates_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        math_helpers.spherical_to_cartesian_coordinates, shape)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (1,)),)
  def test_spherical_to_cartesian_coordinates_raised(self, error_msg, *shape):
    """Tests that the shape exception is raised."""
    self.assert_exception_is_raised(
        math_helpers.spherical_to_cartesian_coordinates, error_msg, shape)

  def test_spherical_to_cartesian_coordinates_jacobian_random(self):
    """Test the Jacobian of the spherical_to_cartesian_coordinates function."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    r_init = np.random.uniform(0.0, 10.0, size=tensor_shape + [1])
    theta_init = np.random.uniform(
        -np.pi / 2.0, np.pi / 2.0, size=tensor_shape + [1])
    phi_init = np.random.uniform(-np.pi, np.pi, size=tensor_shape + [1])
    data_init = np.stack((r_init, theta_init, phi_init), axis=-1)

    self.assert_jacobian_is_correct_fn(
        math_helpers.spherical_to_cartesian_coordinates, [data_init])

  @parameterized.parameters(
      (((0.0, 0.0),), ((1.0, 0.0, 0.0),)),
      (((1.0, 0.0),), ((1.0, np.pi, 0.0),)),
      (((0.0, 1.0),), ((1.0, 0.0, 2.0 * np.pi),)),
      (((1.0, 1.0),), ((1.0, np.pi, 2.0 * np.pi),)),
  )
  def test_square_to_spherical_coordinates_preset(self, test_inputs,
                                                  test_outputs):
    """Tests that square_to_spherical_coordinates generates expected results."""
    self.assert_output_is_correct(math_helpers.square_to_spherical_coordinates,
                                  test_inputs, test_outputs)

  def test_square_to_spherical_coordinates_jacobian_random(self):
    """Tests the Jacobian of square_to_spherical_coordinates."""
    epsilon = 1e-3
    point_2d_init = np.random.uniform(epsilon, 1.0 - epsilon, size=(10, 2))

    self.assert_jacobian_is_correct_fn(
        math_helpers.square_to_spherical_coordinates, [point_2d_init],
        atol=1e-3)

  def test_square_to_spherical_coordinates_range_exception_raised(self):
    """Tests that the exceptions are raised correctly."""
    point_2d_below = np.random.uniform(-1.0, -sys.float_info.epsilon, size=(2,))
    point_2d_above = np.random.uniform(
        1.0 + asserts.select_eps_for_addition(tf.float32), 2.0, size=(2,))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          math_helpers.square_to_spherical_coordinates(point_2d_below))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          math_helpers.square_to_spherical_coordinates(point_2d_above))

  @parameterized.parameters(
      ((2,),),
      ((None, 2),),
  )
  def test_square_to_spherical_coordinates_shape_exception_not_raised(
      self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        math_helpers.square_to_spherical_coordinates, shape)

  @parameterized.parameters(
      ("must have exactly 2 dimensions in axis -1", (1,)),
      ("must have exactly 2 dimensions in axis -1", (3,)),
  )
  def test_square_to_spherical_coordinates_shape_exception_raised(
      self, error_msg, *shape):
    """Tests that the shape exception is raised."""
    self.assert_exception_is_raised(
        math_helpers.square_to_spherical_coordinates, error_msg, shape)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for linear blend skinning."""

# pylint: disable=line-too-long

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import flagsaver
from absl.testing import parameterized
import tensorflow as tf

from tensorflow_graphics.geometry.transformation import linear_blend_skinning
from tensorflow_graphics.geometry.transformation.tests import test_helpers
from tensorflow_graphics.util import test_case


class LinearBlendSkinningTest(test_case.TestCase):

  # pyformat: disable
  @parameterized.parameters(
      ((3,), (7,), (7, 3, 3), (7, 3)),
      ((None, 3), (None, 9), (None, 9, 3, 3), (None, 9, 3)),
      ((7, 1, 3), (1, 4, 11), (5, 11, 3, 3), (1, 11, 3)),
      ((7, 4, 3), (4, 11), (11, 3, 3), (11, 3)),
      ((3,), (5, 4, 11), (11, 3, 3), (11, 3)),
  )
  # pyformat: enable
  def test_blend_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(linear_blend_skinning.blend, shapes)

  # pyformat: disable
  @parameterized.parameters(
      ("points must have exactly 3 dimensions in axis -1",
       (None,), (7,), (7, 3, 3), (7, 3)),
      ("bone_rotations must have a rank greater than 2", (3,), (7,), (3, 3), (3,)),
      ("bone_rotations must have exactly 3 dimensions in axis -1",
       (3,), (7,), (7, 3, None), (7, 3)),
      ("bone_rotations must have exactly 3 dimensions in axis -2",
       (3,), (7,), (7, None, 3), (7, 3)),
      ("bone_translations must have a rank greater than 1", (3,), (7,), (7, 3, 3), (3,)),
      ("bone_translations must have exactly 3 dimensions in axis -1",
       (3,), (7,), (7, 3, 3), (7, None)),
      (r"Tensors \[\'skinning_weights\', \'bone_rotations\'\] must have the same number of dimensions in axes",
       (3,), (9,), (7, 3, 3), (9, 3)),
      (r"Tensors \[\'skinning_weights\', \'bone_translations\'\] must have the same number of dimensions in axes",
       (3,), (9,), (9, 3, 3), (7, 3)),
      ("Not all batch dimensions are broadcast-compatible",
       (2, 3, 3), (3, 1, 7), (7, 3, 3), (7, 3)),
      ("Not all batch dimensions are broadcast-compatible",
       (2, 3, 3), (2, 1, 7), (3, 7, 3, 3), (2, 7, 3)),
  )
  # pyformat: enable
  def test_blend_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(linear_blend_skinning.blend, error_msg,
                                    shapes)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_blend_jacobian_random(self):
    """Test the Jacobian of the blend function."""
    (x_points_init, x_weights_init, x_rotations_init,
     x_translations_init) = test_helpers.generate_random_test_lbs_blend()

    self.assert_jacobian_is_correct_fn(
        linear_blend_skinning.blend,
        [x_points_init, x_weights_init, x_rotations_init, x_translations_init])

  def test_blend_preset(self):
    """Checks that blend returns the expected value."""
    (x_points_init, x_weights_init, x_rotations_init, x_translations_init,
     y_blended_points_init) = test_helpers.generate_preset_test_lbs_blend()

    x_points = tf.convert_to_tensor(value=x_points_init)
    x_weights = tf.convert_to_tensor(value=x_weights_init)
    x_rotations = tf.convert_to_tensor(value=x_rotations_init)
    x_translations = tf.convert_to_tensor(value=x_translations_init)
    y_blended_points = tf.convert_to_tensor(value=y_blended_points_init)

    y = linear_blend_skinning.blend(x_points, x_weights, x_rotations,
                                    x_translations)

    self.assertAllClose(y_blended_points, y)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data augmentation utility functions."""

import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.transformation import rotation_matrix_3d


def jitter(points, stddev=0.01, clip=0.05):
  """Randomly jitters points with (clipped) Gaussian noise."""
  assert (stddev > 0), "jitter needs a positive sigma"
  assert (clip > 0), "jitter needs a positive clip"
  noise = tf.random.normal(points.shape, stddev=stddev)
  noise = tf.clip_by_value(noise, -clip, +clip)
  return points + noise


def rotate(points):
  """Randomly rotates a point cloud around the Y axis (UP)."""
  axis = tf.constant([[0, 1, 0]], dtype=tf.float32)  # [1, 3]
  angle = tf.random.uniform((1, 1), minval=0., maxval=2 * np.pi)  # [1, 1]
  matrix = rotation_matrix_3d.from_axis_angle(axis, angle)  # [3, 3]
  return rotation_matrix_3d.rotate(points, matrix)  # [N, 3]



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the graph convolution layers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

import tensorflow_graphics.nn.layer.graph_convolution as gc_layer
from tensorflow_graphics.util import test_case


def _dense_to_sparse(data):
  """Convert a numpy array to a tf.SparseTensor."""
  indices = np.where(data)
  return tf.SparseTensor(
      np.stack(indices, axis=-1), data[indices], dense_shape=data.shape)


def _dummy_data(batch_size, num_vertices, num_channels):
  """Create inputs for feature_steered_convolution."""
  if batch_size > 0:
    data = np.zeros(
        shape=(batch_size, num_vertices, num_channels), dtype=np.float32)
    neighbors = _dense_to_sparse(
        np.tile(np.eye(num_vertices, dtype=np.float32), (batch_size, 1, 1)))
  else:
    data = np.zeros(shape=(num_vertices, num_channels), dtype=np.float32)
    neighbors = _dense_to_sparse(np.eye(num_vertices, dtype=np.float32))
  return data, neighbors


class GraphConvolutionTestFeatureSteeredConvolutionLayerTests(
    test_case.TestCase):

  @parameterized.parameters(
      (1, 1, 1, 1, 1, False),
      (4, 2, 3, None, 5, False),
      (1, 2, 3, 4, 5, True),
  )
  def test_feature_steered_convolution_layer_exception_not_raised_shapes(
      self, batch_size, num_vertices, in_channels, out_channels,
      num_weight_matrices, translation_invariant):
    """Check if the convolution parameters and output have correct shapes."""
    data, neighbors = _dummy_data(batch_size, num_vertices, in_channels)
    name_scope = "test"
    if tf.executing_eagerly():
      layer = gc_layer.FeatureSteeredConvolutionKerasLayer(
          translation_invariant=translation_invariant,
          num_weight_matrices=num_weight_matrices,
          num_output_channels=out_channels,
          name=name_scope)

    def _run_convolution():
      """Run the appropriate feature steered convolution layer."""
      if tf.executing_eagerly():
        try:
          output = layer(inputs=[data, neighbors], sizes=None)
        except Exception as e:  # pylint: disable=broad-except
          self.fail("Exception raised: %s" % str(e))
      else:
        try:
          output = gc_layer.feature_steered_convolution_layer(
              data=data,
              neighbors=neighbors,
              sizes=None,
              translation_invariant=translation_invariant,
              num_weight_matrices=num_weight_matrices,
              num_output_channels=out_channels,
              name=None,
              var_name=name_scope)
        except Exception as e:  # pylint: disable=broad-except
          self.fail("Exception raised: %s" % str(e))
      return output

    output = _run_convolution()
    output_shape = output.shape.as_list()
    out_channels = in_channels if out_channels is None else out_channels

    self.assertEqual(output_shape[-1], out_channels)
    self.assertAllEqual(output_shape[:-1], data.shape[:-1])

    def _get_var_shape(var_name):
      """Get the shape of a variable by name."""
      if tf.executing_eagerly():
        trainable_variables = layer.trainable_variables
        for tv in trainable_variables:
          if tv.name == name_scope + "/" + var_name + ":0":
            return tv.shape.as_list()
        raise ValueError("Variable not found.")
      else:
        with tf.compat.v1.variable_scope(name_scope, reuse=True):
          variable = tf.compat.v1.get_variable(
              var_name, initializer=tf.constant(0))
          return variable.shape.as_list()

    self.assertAllEqual(_get_var_shape("u"), [in_channels, num_weight_matrices])
    self.assertAllEqual(_get_var_shape("c"), [num_weight_matrices])
    self.assertAllEqual(_get_var_shape("b"), [out_channels])
    self.assertAllEqual(
        _get_var_shape("w"), [num_weight_matrices, in_channels, out_channels])
    if not translation_invariant:
      self.assertAllEqual(
          _get_var_shape("v"), [in_channels, num_weight_matrices])

  def test_feature_steered_convolution_layer_initializer(self):
    """Tests a custom variable initializer."""
    data = np.array(((1.0, 1.0), (-1.0, 1.0), (-1.0, -1.0), (1.0, -1.0)))
    neighbors_indices = np.array(((0, 0), (0, 1), (0, 3),
                                  (1, 0), (1, 1), (1, 2),
                                  (2, 1), (2, 2), (2, 3),
                                  (3, 0), (3, 2), (3, 3)))
    neighbors = tf.SparseTensor(
        neighbors_indices, np.ones(shape=(12,)) / 3.0, dense_shape=(4, 4))
    initializer = tf.compat.v1.keras.initializers.zeros()

    if tf.executing_eagerly():
      layer = gc_layer.FeatureSteeredConvolutionKerasLayer(
          translation_invariant=False,
          initializer=initializer)
      output = layer(inputs=[data, neighbors], sizes=None)
    else:
      out = gc_layer.feature_steered_convolution_layer(
          data=data,
          neighbors=neighbors,
          sizes=None,
          translation_invariant=False,
          initializer=initializer)
      self.evaluate(tf.compat.v1.global_variables_initializer())
      output = self.evaluate(out)

    # All zeros initializer should result in all zeros output.
    self.assertAllEqual(output, np.zeros_like(data))

  def test_feature_steered_convolution_layer_training(self):
    """Test a simple training loop."""
    # Generate a small valid input for a simple training task.
    # Four corners of a square.
    data = np.array(((1.0, 1.0), (-1.0, 1.0), (-1.0, -1.0), (1.0, -1.0)))
    neighbors_indices = np.array(((0, 0), (0, 1), (0, 3),
                                  (1, 0), (1, 1), (1, 2),
                                  (2, 1), (2, 2), (2, 3),
                                  (3, 0), (3, 2), (3, 3)))
    neighbors = tf.SparseTensor(
        neighbors_indices, np.ones(shape=(12,)) / 3.0, dense_shape=(4, 4))
    # Desired output is arbitrary.
    labels = np.reshape([-1.0, -0.5, 0.5, 1.0], (-1, 1))
    num_training_iterations = 5

    if tf.executing_eagerly():
      with tf.GradientTape(persistent=True) as tape:
        layer = gc_layer.FeatureSteeredConvolutionKerasLayer(
            translation_invariant=False,
            num_weight_matrices=1,
            num_output_channels=1)
        output = layer(inputs=[data, neighbors], sizes=None)
        loss = tf.nn.l2_loss(output - labels)

      trainable_variables = layer.trainable_variables
      for _ in range(num_training_iterations):
        grads = tape.gradient(loss, trainable_variables)
        tf.compat.v1.train.GradientDescentOptimizer(1e-4).apply_gradients(
            zip(grads, trainable_variables))
    else:
      output = gc_layer.feature_steered_convolution_layer(
          data=data,
          neighbors=neighbors,
          sizes=None,
          translation_invariant=False,
          num_weight_matrices=1,
          num_output_channels=1)
      train_op = tf.compat.v1.train.GradientDescentOptimizer(1e-4).minimize(
          tf.nn.l2_loss(output - labels))
      with tf.compat.v1.Session() as sess:
        sess.run(tf.compat.v1.initialize_all_variables())
        for _ in range(num_training_iterations):
          sess.run(train_op)


class GraphConvolutionTestDynamicGraphConvolutionKerasLayerTests(
    test_case.TestCase):

  @parameterized.parameters(
      (1, 1, 1, 1, "weighted"),
      (4, 2, 3, 12, "max"),
      (1, 2, 3, 4, "max"),
  )
  def test_dynamic_graph_convolution_keras_layer_exception_not_raised_shapes(
      self, batch_size, num_vertices, in_channels, out_channels, reduction):
    """Check if the convolution parameters and output have correct shapes."""
    if not tf.executing_eagerly():
      return
    data, neighbors = _dummy_data(batch_size, num_vertices, in_channels)
    layer = gc_layer.DynamicGraphConvolutionKerasLayer(
        num_output_channels=out_channels,
        reduction=reduction)

    try:
      output = layer(inputs=[data, neighbors], sizes=None)
    except Exception as e:  # pylint: disable=broad-except
      self.fail("Exception raised: %s" % str(e))

    self.assertAllEqual((batch_size, num_vertices, out_channels), output.shape)

  @parameterized.parameters(
      (1, 1, 1, 1, "weighted"),
      (4, 2, 3, 12, "max"),
      (1, 2, 3, 4, "max"),
  )
  def test_dynamic_graph_convolution_keras_layer_zero_kernel(
      self, batch_size, num_vertices, in_channels, out_channels, reduction):
    """Tests convolution with an all-zeros kernel."""
    if not tf.executing_eagerly():
      return
    data, neighbors = _dummy_data(batch_size, num_vertices, in_channels)
    data = np.random.uniform(size=data.shape).astype(np.float32)
    layer = gc_layer.DynamicGraphConvolutionKerasLayer(
        num_output_channels=out_channels,
        reduction=reduction,
        use_bias=False,
        kernel_initializer=tf.compat.v1.keras.initializers.zeros())

    output = layer(inputs=[data, neighbors], sizes=None)

    self.assertAllEqual(
        output,
        np.zeros(shape=(batch_size, num_vertices, out_channels),
                 dtype=np.float32))

  @parameterized.parameters((1, 1, 1), (2, 3, 12), (2, 3, 4))
  def test_dynamic_graph_convolution_keras_layer_duplicate_features(
      self, num_vertices, in_channels, out_channels):
    """Tests convolution when all vertex features are identical."""
    if not tf.executing_eagerly():
      return
    data = np.random.uniform(size=(1, in_channels))
    data = np.tile(data, (num_vertices, 1))
    # Results should be independent of 'neighbors'.
    neighbors = np.maximum(np.random.randint(
        0, 2, size=(num_vertices, num_vertices)), np.eye(num_vertices))
    neighbors = _dense_to_sparse(neighbors)
    layer = gc_layer.DynamicGraphConvolutionKerasLayer(
        num_output_channels=out_channels,
        reduction="max")

    output = layer(inputs=[data, neighbors], sizes=None)

    output_tile = tf.tile(output[:1, :], (num_vertices, 1))

    self.assertAllEqual(output, output_tile)

  @parameterized.parameters("weighted", "max")
  def test_dynamic_graph_convolution_keras_layer_training(self, reduction):
    """Test a simple training loop."""
    if not tf.executing_eagerly():
      return
    # Generate a small valid input for a simple training task.
    # Four corners of a square.
    data = np.array(((1.0, 1.0), (-1.0, 1.0), (-1.0, -1.0), (1.0, -1.0)))
    neighbors_indices = np.array(((0, 0), (0, 1), (0, 3),
                                  (1, 0), (1, 1), (1, 2),
                                  (2, 1), (2, 2), (2, 3),
                                  (3, 0), (3, 2), (3, 3)))
    neighbors = tf.SparseTensor(
        neighbors_indices, np.ones(shape=(12,)) / 3.0, dense_shape=(4, 4))
    # Desired output is arbitrary.
    labels = np.reshape([-1.0, -0.5, 0.5, 1.0], (-1, 1))
    num_training_iterations = 5

    with tf.GradientTape(persistent=True) as tape:
      layer = gc_layer.DynamicGraphConvolutionKerasLayer(
          num_output_channels=2,
          reduction=reduction)
      output = layer(inputs=[data, neighbors], sizes=None)
      loss = tf.nn.l2_loss(output - labels)

      trainable_variables = layer.trainable_variables
      for _ in range(num_training_iterations):
        grads = tape.gradient(loss, trainable_variables)
        tf.compat.v1.train.GradientDescentOptimizer(1e-4).apply_gradients(
            zip(grads, trainable_variables))

if __name__ == "__main__":
  test_case.main()
